#pragma once

#include <assert.h>

#include "Vector.h"

#if !defined(_WIN32)
#define __forceinline
#define ASSERT(x) assert(x)
#endif

__forceinline
void closest_point_vertex ( Vector3& v0,	// Face vertex v0.
							Vector3& p,		// point
							float * d,		// Current closest distance
							Vector3& pc )	// Current closest point
{
	Vector3 diff = p - v0;

	double length_squared = diff.Dot( diff );

	// save a square root operation if we fail
	// at the expense of an unconditional multiplication
	if ( length_squared > (*d)*(*d) ) 
		return;

	*d = (float)sqrt( length_squared );
	pc = v0;
}

inline
void closest_point_edge (	Vector3& v0,		// Edge vertex v0.
							Vector3& v1,		// Edge vertex v1.
							Vector3& p,			// point
							float * d,		// Current closest distance
							Vector3& pc )	// Current closest point
{
	Vector3 dir = v1 - v0;

	double dot1 = dir.Dot( p-v0 );
	double dot2 = dir.Dot( dir );

	double t = dot1 / dot2;

	Vector3 v;

	// v1 is the closest point
	if ( t > 1.f )
	{
		v = v1;
	}
	// v0 is the closest point
	else if ( t < 0 )
	{
		v = v0;		
	}
	else
	{
		v = v0 + dir*t;
	}

	Vector3 diff = p - v;
	double length_squared = diff.Dot( diff );

	// save a square root operation if we fail
	// at the expense of an unconditional multiplication
	if ( length_squared > (*d)*(*d) ) 
		return;

	*d = (float)sqrt( length_squared );
	pc = v;
}

inline
void closest_point_triangle ( Vector3& v0,		// Face vertex v0.
							Vector3& v1,		// Face vertex v1.
							Vector3& v2,		// Face vertex v2.
							Vector3& p,			// point
							float * d,		// Current closest distance
							Vector3& pc )	// Current closest point
{
	Vector3 edge1 = v1-v0;
	Vector3 edge2 = v2-v0;
	Vector3 edge3 = v2-v1;

	Vector3 n = edge1.Cross( edge2 ).Normalize();
	Vector3 dir = p - v0;

	double dot_dn = dir.Dot( n );

	// Project the point p into the plane

	Vector3 v = v0 + dir - n*dot_dn;

	// To simplify barycentric co-ord calc to 2D, project to primary
	// axis plane of closest orientation.
	int ip=0;
	double n_max = (double)fabs(n[0]);
	if ( (double)fabs(n[1]) > n_max ) { n_max = (double)fabs(n[1]); ip = 1; }
	if ( (double)fabs(n[2]) > n_max ) { ip = 2; }

	double x1[2], x2[2], x3[2], x0[2];
	int j=0;
	for ( int i=0 ; i<3 ; i++ )
	{
		if ( i == ip ) continue;
		x1[j] = v0[i];
		x2[j] = v1[i];
		x3[j] = v2[i];
		x0[j] = v[i];
		j++;
	}

	// 2D Barycentric coordinates.
	double b0 =  (x2[0] - x1[0]) * (x3[1] - x1[1]) - (x3[0] - x1[0]) * (x2[1] - x1[1]);
	double b1 = ((x2[0] - x0[0]) * (x3[1] - x0[1]) - (x3[0] - x0[0]) * (x2[1] - x0[1])) / b0;
	double b2 = ((x3[0] - x0[0]) * (x1[1] - x0[1]) - (x1[0] - x0[0]) * (x3[1] - x0[1])) / b0;
	double b3 = ((x1[0] - x0[0]) * (x2[1] - x0[1]) - (x2[0] - x0[0]) * (x1[1] - x0[1])) / b0;

	if( b1 < 0 || b2 < 0 || b3 < 0 || b1 > 1 || b2 > 1 || b3 > 1 ) 
		return;

	Vector3 diff = p - v;

	double length_squared = diff.Dot( diff );

	// save a square root operation if we fail
	// at the expense of an unconditional multiplication
	if ( length_squared > (*d)*(*d) ) 
		return;

	*d = (float)sqrt( length_squared );
	pc = v;
}

__forceinline
bool point_inside_cone (
	Vector3& p,		// Origin of the cone
	Vector3& dir,	// Direction of the cone (normalized)
	double  cos_t,	// Cosine of the angle defining the cone
	Vector3& v )	// Point to be tested
{
	Vector3 diff = v-p;

	double length_squared	= diff.Dot( diff );
	double dot				= diff.Dot( dir );

	// this is the same thing as dot^2 > ( cos_t * length )^2
	if ( dot > 0 && dot*dot > cos_t * cos_t * length_squared ) 
	{
		return true;
	}

  	return false;
}

__forceinline
bool line_outside_cone (
	Vector3& p,		// Origin of the cone
	Vector3& dir,	// Direction of the cone (normalized)
	double cos_t,	// Cosine of the angle defining the cone
	Vector3& v0,	// Line to be tested v0, v1
	Vector3& v1)		
{
	if ( point_inside_cone ( p, dir, cos_t, v0 ) ) return false;
	if ( point_inside_cone ( p, dir, cos_t, v1 ) ) return false;
	return true;
}

__forceinline
bool triangle_outside_cone (
	Vector3& p,		// Origin of the cone
	Vector3& dir,	// Direction of the cone (normalized)
	double cos_t,	// Cosine of the angle defining the cone
	Vector3& v0,		// Face vertex 0
	Vector3& v1,		// Face vertex 1
	Vector3& v2 )		// Face vertex 2
{
	if ( point_inside_cone ( p, dir, cos_t, v0 ) ) return false;
	if ( point_inside_cone ( p, dir, cos_t, v1 ) ) return false;
	if ( point_inside_cone ( p, dir, cos_t, v2 ) ) return false;
	return true;
}

__forceinline bool sphere_intersects_cone(	Vector3&	vSphereCent,
									float&		fSphereRad,
									Vector3&	vConeVertex,
									Vector3&	vConeDir,
									float		fConeAngle
								)
{
	Vector3 u = vConeVertex - vConeDir*( fSphereRad / sin( fConeAngle ) );
	Vector3 d = vSphereCent - u;

	// avoid a square root and compare the squares instead
	// we're essentially comparing ( scalar projection of d to the viewing dir > length of d )
	double dot = vConeDir.Dot( d );
	double cos_t = cos( fConeAngle );
	return ( dot > 0 && dot*dot >= d.Dot( d ) * cos_t * cos_t);
}

inline
bool clip_edge_with_cone( Vector3& vE0, Vector3& vE1, Vector3& vOrg, Vector3& vAxis, float fCosAngle )
{
	bool bOneVertOk = true;
	
	if ( !point_inside_cone( vOrg, vAxis, fCosAngle, vE0 ) )
	{
		bOneVertOk = false;
	}
	
	if ( !point_inside_cone( vOrg, vAxis, fCosAngle, vE1 ) )
	{
		if ( !bOneVertOk )
		{

			return false;
		}
	}

	return true;
}

inline 
bool clip_edge_with_plane( Vector3& vE0, Vector3& vE1, Vector3& vClipPlaneNorm, float fClipPlaneConst )
{
	double d0 = vClipPlaneNorm.Dot( vE0 ) + fClipPlaneConst;
	double d1 = vClipPlaneNorm.Dot( vE1 ) + fClipPlaneConst;

	// Trivial pass: both vertices on the right side
	if ( d0 >= 0 && d1 >= 0 )
		return true;

	// Trivial fail: both vertices on the wrong side
	if ( d0 < 0 && d1 < 0 )
		return false;

	// Vertex vE0 is on the right side, the other on the wrong side
	if ( d0 > 0 )
	{
		Vector3 vRay = vE0 - vE1;
		d0	= vClipPlaneNorm.Dot( vRay );	
		assert( d0 != 0 );

		d1	= -(vClipPlaneNorm.Dot( vE1 ) + fClipPlaneConst);
		vE1	= vE1 + vRay*(d1/d0);
	}

	// Vertex vE1 is on the right side, the other on the wrong side
	else
	{
		Vector3 vRay = vE1 - vE0;
		d0	= vClipPlaneNorm.Dot( vRay );	
		assert( d0 != 0 );

		d1	= -(vClipPlaneNorm.Dot( vE0 ) + fClipPlaneConst);
		vE0	= vE0 + vRay*(d1/d0);
	}

	return true;
}

inline
bool ray_intersects_triangle( Vector3& v0, Vector3& v1, Vector3& v2, Vector3& vP, Vector3& vViewDir, double& t )
{
	Vector3 vE1, vE2, tvec, pvec, qvec;
	double det,inv_det;
	double u,v;

	// find vectors for two edges sharing triangle[0] 
	vE1	= v1 - v0;
	vE2	= v2 - v0;

	// begin calculating determinant - also used to calculate U parameter 
	pvec	= vViewDir.Cross( vE2 );

	// if determinant is near zero, ray lies in plane of triangle
	det		= vE1.Dot( pvec );

	if (det > -EPSILON && det < EPSILON)
		return false;

	inv_det = 1.0 / det;

	// calculate distance from v0 to ray origin 
	tvec	= vP - v0;

	// calculate U parameter and test bounds 
	u		= tvec.Dot( pvec ) * inv_det;
	if ( u < 0.0 || u > 1.0)
		return false;

	// prepare to test V parameter 
	qvec	= tvec.Cross( vE1 );

	// calculate V parameter and test bounds 
	v = vViewDir.Dot( qvec ) * inv_det;
	if ( v < 0.0 || u + v > 1.0 )
		return false;

	// calculate t, ray intersects triangle 
	t		= vE2.Dot( qvec ) * inv_det;

	if ( t < 0.0 )
		return false;

	return true;
}

#if !defined(_WIN32)
#undef __forceinline
#undef ASSERT(x)
#endif

