#include "Histogram.h"
#include "assert.h"
#include <cmath>
//#include "WM3Math.h"

#define M_PI 3.14159

Histogram::Histogram( int bucketCount, float minValue, float maxValue )
{
	m_bucketCount = bucketCount;
	m_minValue = minValue;
	m_maxValue = maxValue;
	m_bucketFrequency = new int[ bucketCount ];

	highestBucket=0;

	/*
	for ( int i = 0; i < MIX_MODEL_COUNT; i++ )
	{
		m_bucketMixWeight[i] = new float[ bucketCount ];
	}
	*/

	Clear();
	ResetMixtureModel();
}

Histogram::~Histogram()
{
	delete [] m_bucketFrequency;

	/*
	for ( int i = 0; i < MIX_MODEL_COUNT; i++ )
	{
		delete [] m_bucketMixWeight[i];
	}
	*/
}

int Histogram::GetBucketCount() const
{ return m_bucketCount; }

int Histogram::GetValueCount() const
{ return m_valueCount; }

float Histogram::GetMinValue() const
{ return m_minValue; }

float Histogram::GetMaxValue() const
{ return m_maxValue; }

float Histogram::GetValue( int bucketIndex ) const
{
	int bucketIndexMax = m_bucketCount - 1;
	float value = (float)bucketIndex / (float)bucketIndexMax;

	value = value * ( m_maxValue - m_minValue) + m_minValue;
	return value;
}

int Histogram::GetMaxFrequency() const
{ return m_maxFrequency; }

int Histogram::GetFrequency( int bucketIndex ) const
{
	assert( bucketIndex >= 0 && bucketIndex < m_bucketCount );

	return m_bucketFrequency[ bucketIndex ];
}

float Histogram::GetMedian() const
{
	int frequencyTotal = 0;
	for ( int i = 0; i < m_bucketCount; i++ )
	{
		frequencyTotal += m_bucketFrequency[i];

		if ( frequencyTotal >= m_valueCount / 2 )
		{ return GetValue(i); }
	}

	return GetValue( m_bucketCount / 2);
}

float Histogram::GetMean() const
{
	float valueTotal = 0.0;
	int countTotal = 0;

	for ( int i = 0; i < m_bucketCount; i++ )
	{
		valueTotal += GetValue(i) * (float)m_bucketFrequency[i];
		countTotal += m_bucketFrequency[i];
	}

	return valueTotal / (float)countTotal;
}

float Histogram::GetTukeyMean() const
{
	// perform Tukey's biweight algorithm to find a robust mean
	//
	// algorithm is described in Appendix I of the following document:
	// http://www.affymetrix.com/support/technical/whitepapers/sadd_whitepaper.pdf

	float M = GetMedian();

	Histogram medianDistHistogram = Histogram( m_bucketCount, m_minValue, m_maxValue );

	for ( int i = 0; i < m_bucketCount; i++ )
	{
		float x = GetValue(i);
		medianDistHistogram.AddValue( fabs( x - M ) );
	}

	float S = medianDistHistogram.GetMedian();

	const float c = 5.0;
	const float e = 0.0001f;

	float T = 0.0;
	float wT = 0.0;

	for ( int i = 0; i < m_bucketCount; i++ )
	{
		float x = GetValue(i);
		float u = ( x - M ) / ( c * S + e );
		float wu = ( fabs(u) <= 1.0f ) ? (1.0f - u*u) * ( 1.0f - u*u) : 0.0f;

		T += wu * x;
		wT += wu;
	}

	//T = ( wT > DBL_EPSILON ) ? ( T / wT ) : 1.0;
	T = ( wT > 1e-10 ) ? ( T / wT ) : 1.0f;

	return T;
}

/*
void Histogram::SetValueBound( float minValue, float maxValue )
{
	m_minValue = minValue;
	m_maxValue = maxValue;
}
*/

void Histogram::AddValue( float value )
{
	value = ( value - m_minValue ) / ( m_maxValue - m_minValue );

	int bucketIndexMin = 0;
	int bucketIndexMax = m_bucketCount - 1;
	int bucketIndex = (int)( (float)bucketIndexMax * value );

	if ( bucketIndex < bucketIndexMin ) { bucketIndex = bucketIndexMin; }
	if ( bucketIndex > bucketIndexMax ) { bucketIndex = bucketIndexMax; }

	if (bucketIndex>highestBucket)
		highestBucket=bucketIndex;

	m_bucketFrequency[ bucketIndex ]++;

	if ( m_bucketFrequency[ bucketIndex ] > m_maxFrequency )
	{
		m_maxFrequency = m_bucketFrequency[ bucketIndex ];
		highestBucketIndex=bucketIndex;
	}

	m_valueCount++;
}	

void Histogram::Clear()
{
	for ( int i = 0; i < m_bucketCount; i++ )
	{
		m_bucketFrequency[i] = 0;
	}

	m_valueCount = 0;
	m_maxFrequency = -1;
}

void Histogram::FitMixtureOfGaussians()
{
	// perform EM algorithm to find the best parameters for the mixture model

	// TODO: make this a member variable to prevent memory allocation on every function call
	double *PixcArray = new double [m_bucketCount];

	for ( int t = 0; t < 10; t++ )
	{
		for ( int i = 0; i < MIX_MODEL_COUNT; i++ )
		{
			double Pi = m_mixModel[i].weight;
			double ui = m_mixModel[i].mean;
			double si = m_mixModel[i].std;

			double PiNew = 0.0;
			double uiNew = 0.0;
			double siNew = 0.0;
			double sumPixc = 0.0;
			
			for ( int c = 0; c < m_bucketCount; c++ )
			{
				double xc = (double)GetFrequency(c) / (double)GetMaxFrequency();
				double Pxci = Get1DGaussianSample( xc, ui, si );

				double Pxc = 0.0;
				for ( int j = 0; j < MIX_MODEL_COUNT; j++ )
				{
					double Pj = m_mixModel[j].weight;
					double uj = m_mixModel[j].mean;
					double sj = m_mixModel[j].std;

					double Pxcj = Get1DGaussianSample( xc, uj, sj );
					Pxc += Pj * Pxcj;
				}

				//double Pixc = ( Pxc < DBL_EPSILON ) ? 1.0 : ( Pi * Pxci ) / Pxc;
				double Pixc = ( Pxc < 1e-10 ) ? 1.0 : ( Pi * Pxci ) / Pxc;
				PixcArray[c] = Pixc;

				sumPixc += Pixc;
				uiNew += Pixc * xc;
			}

			PiNew = sumPixc / (double)m_bucketCount;
			uiNew = uiNew / sumPixc;

			for ( int c = 0; c < m_bucketCount; c++ )
			{
				double xc = (double)GetFrequency(c) / (double)GetMaxFrequency();
				double Pixc = PixcArray[c];

				siNew += Pixc * ( xc - uiNew )*( xc - uiNew );
			}

			siNew = sqrt( siNew / sumPixc );

			m_mixModel[i].mean = (float)uiNew;
			m_mixModel[i].std = (float)siNew;
			m_mixModel[i].weight = (float)PiNew;
		}
	}

	delete [] PixcArray;
}

void Histogram::ResetMixtureModel()
{
	for ( int i = 0; i < MIX_MODEL_COUNT; i++ )
	{
		float mean = (float)i / (float)(MIX_MODEL_COUNT - 1) * 0.5f + 0.25f;
		mean = ( mean - m_minValue ) / ( m_maxValue - m_minValue );
		
		m_mixModel[i].mean = mean;
		m_mixModel[i].std = 0.125f / (float)(MIX_MODEL_COUNT - 1);
		m_mixModel[i].weight = 1.0f / (float)MIX_MODEL_COUNT;

		/*
		for ( int j = 0; j < m_bucketCount; j++ )
		{
			m_bucketMixWeight[i][j] = 1.0f / (float)(m_bucketCount);
		}
		*/
	}
}

Gaussian Histogram::GetMixModel( int modelIndex ) const
{
	assert( modelIndex >= 0 && modelIndex < MIX_MODEL_COUNT );

	return m_mixModel[modelIndex];
}

double Histogram::Get1DGaussianSample( double x, double mean, double std )
{
	static const double sqrt2PI = sqrt( 2.0 * 2.0 * M_PI );

	double dist = x - mean;
	return exp( -(dist * dist) / ( 2.0 * std * std ) ) / ( std *  sqrt2PI );	
}

float Histogram::GetHighestBucketValue() {

	return ((float)highestBucket/(float)m_bucketCount)*(m_maxValue-m_minValue)+m_minValue;

}

float Histogram::GetValleyAfterPeakValue() {
	
	int curIndex=highestBucketIndex+1;

	while (curIndex+2<m_bucketCount&& (m_bucketFrequency[curIndex+1]<m_bucketFrequency[curIndex]||
										m_bucketFrequency[curIndex+2]<m_bucketFrequency[curIndex]))
		curIndex++;

	return ((float)curIndex/(float)m_bucketCount)*(m_maxValue-m_minValue)+m_minValue;

}