#include "Mesh.h"

Mesh::Mesh()
{
    drawSimpleDispList=0;

    boundingSphereRad=1.0;
    boundingSphereCentre=Vector3(0,1,0);
    meshLoBounds=Vector3(-1,0,-1);
    meshHiBounds=Vector3(1,2,1);

    bitmapData=NULL;
    bitmapLData=NULL;
}

void Mesh::SetVertsAndFaces(const QVector <Vector3> & v, const QVector <QVector <int> > & f) {

    m_vVertices.clear();
    m_vFaces.clear();
    triangleBounds.clear();

    m_vVertices=v;
    m_vFaces=f;

    FillInInfo();

}

void Mesh::FlipFaceNorms() {

    for (int i=0; i<m_vFaceNorms.size(); ++i)
        m_vFaceNorms[i]=m_vFaceNorms[i]*(-1.0f);

}

void Mesh::FillInInfo() {

    // Process face normals
    m_vFaceNorms.clear();
    for(int nFace=0; nFace<m_vFaces.size(); ++nFace)
    {
        int vertIndex;
        float triangle[3][3];

        for(int whichVertex = 0; whichVertex < m_vFaces[nFace].size(); ++whichVertex)
        {
            // Get the vertex index for each point of the face
            vertIndex = m_vFaces[nFace][whichVertex];

            triangle[whichVertex][0] = m_vVertices[vertIndex].x;
            triangle[whichVertex][1] = m_vVertices[vertIndex].y;
            triangle[whichVertex][2] = m_vVertices[vertIndex].z;
        }

        float edge1[3], edge2[3];
        edge1[0] = triangle[0][0] - triangle[1][0];
        edge1[1] = triangle[0][1] - triangle[1][1];
        edge1[2] = triangle[0][2] - triangle[1][2];
        edge2[0] = triangle[2][0] - triangle[1][0];
        edge2[1] = triangle[2][1] - triangle[1][1];
        edge2[2] = triangle[2][2] - triangle[1][2];

        Vector3 eachNorm;
        eachNorm.x = (edge1[1]*edge2[2])-(edge1[2]*edge2[1]);
        eachNorm.y = (edge1[2]*edge2[0])-(edge1[0]*edge2[2]);
        eachNorm.z = (edge1[0]*edge2[1])-(edge1[1]*edge2[0]);
        eachNorm.Normalize();

        m_vFaceNorms.push_back(eachNorm);
    }

    // Get mesh boundaries
    meshLoBounds=Vector3(FLT_MAX,FLT_MAX,FLT_MAX);
    meshHiBounds=Vector3(-FLT_MAX,-FLT_MAX,-FLT_MAX);

    for (int i=0; i<m_vVertices.size(); ++i) {

        meshLoBounds.x=std::min(m_vVertices[i].x,meshLoBounds.x);
        meshLoBounds.y=std::min(m_vVertices[i].y,meshLoBounds.y);
        meshLoBounds.z=std::min(m_vVertices[i].z,meshLoBounds.z);

        meshHiBounds.x=std::max(m_vVertices[i].x,meshHiBounds.x);
        meshHiBounds.y=std::max(m_vVertices[i].y,meshHiBounds.y);
        meshHiBounds.z=std::max(m_vVertices[i].z,meshHiBounds.z);

    }

    boundingSphereCentre=meshLoBounds+(meshHiBounds-meshLoBounds)*0.5f;
    boundingSphereRad=(meshHiBounds-meshLoBounds).Length()*0.5f;

    //normalize mesh
    //TODO: account for fact detected features move around
    //(since file loaded separately for some of them, transformation not applied)

    Vector3 vdiff = meshHiBounds - meshLoBounds;
    float diff = std::max(vdiff.x, std::max(vdiff.y, vdiff.z));

    for (int i=0; i<m_vVertices.size(); ++i) {

        m_vVertices[i].x = (m_vVertices[i].x - boundingSphereCentre.x) / diff;
        m_vVertices[i].y = (m_vVertices[i].y - boundingSphereCentre.y) / diff;
        m_vVertices[i].z = (m_vVertices[i].z - boundingSphereCentre.z) / diff;

    }

    boundingSphereCentre = Vector3(0, 0, 0);
    boundingSphereRad = 0.5f;


    //fill in vertex neighbour info
    for (int i=0; i<m_vVertNeigh.size(); ++i)
        m_vVertNeigh[i].clear();
    m_vVertNeigh.clear();

    for (int i=0; i<m_vVertices.size(); ++i) {
        QSet <int> eachNeigh;
        m_vVertNeigh.push_back(eachNeigh);
    }

    for (int i=0; i<m_vFaces.size(); ++i) {

        int vert1, vert2, vert3;

        vert1=m_vFaces[i][0];
        vert2=m_vFaces[i][1];
        vert3=m_vFaces[i][2];

        m_vVertNeigh[vert1].insert(vert2);
        m_vVertNeigh[vert1].insert(vert3);

        m_vVertNeigh[vert2].insert(vert1);
        m_vVertNeigh[vert2].insert(vert3);

        m_vVertNeigh[vert3].insert(vert1);
        m_vVertNeigh[vert3].insert(vert2);

    }

    //delete former display list (if present)
    if (drawSimpleDispList>0) {
        glDeleteLists(drawSimpleDispList,1);
        drawSimpleDispList=0;
    }

}

float Mesh::GetBoundRad() {
	return boundingSphereRad;
}

Vector3 Mesh::GetBoundLo() {
	return meshLoBounds;
}

Vector3 Mesh::GetBoundHi() {
	return meshHiBounds;
}

Vector3 Mesh::GetBoundCentre() {
	return boundingSphereCentre;
}

void Mesh::DrawSimple() {

    glBegin(GL_TRIANGLES);
    for (int i=0; i<m_vFaces.size(); ++i) {

        glNormal3f(m_vFaceNorms[i].x,m_vFaceNorms[i].y,m_vFaceNorms[i].z);

        for (int j=0; j<3; ++j) {
            glVertex3f(m_vVertices[m_vFaces[i][j]].x,
                       m_vVertices[m_vFaces[i][j]].y,
                       m_vVertices[m_vFaces[i][j]].z);
        }

    }
    glEnd();

}

QVector <Vector3> & Mesh::getVerts() {
        return m_vVertices;
}

QVector <QVector <int> > & Mesh::getFaces() {
        return m_vFaces;
}

QVector <Vector3> & Mesh::getFaceNorms() {
        return m_vFaceNorms;
}

QVector <QSet <int> > & Mesh::getVertNeigh() {
        return m_vVertNeigh;
}

void Mesh::SetTriangleBounds(QVector <Vector3> b) {
	triangleBounds=b;
}

QVector <Vector3> & Mesh::getTriangleBounds() {
	return triangleBounds;
}

void Mesh::SetBitmapData(int *data, int w, int h) {

    //compute laplacian for data array
    if (bitmapData!=NULL)
        delete [] bitmapData;
    if (bitmapLData!=NULL)
        delete [] bitmapLData;

    bitmapData=data;
    bitmapWidth=w;
    bitmapHeight=h;

    bitmapLData=new int[bitmapWidth*bitmapHeight];

    for (int y=0;y<bitmapHeight;y++)
        for (int x=0;x<bitmapWidth;x++) {

        bitmapLData[y*bitmapWidth+x]=0;
        int numSample=0;

        if (x>0) {
            bitmapLData[y*bitmapWidth+x]+=bitmapData[y*bitmapWidth+(x-1)];
            numSample++;
        }

        if (x<bitmapWidth-1) {
            bitmapLData[y*bitmapWidth+x]+=bitmapData[y*bitmapWidth+(x+1)];
            numSample++;
        }

        if (y>0) {
            bitmapLData[y*bitmapWidth+x]+=bitmapData[(y-1)*bitmapWidth+x];
            numSample++;
        }

        if (y<bitmapHeight-1) {
            bitmapLData[y*bitmapWidth+x]+=bitmapData[(y+1)*bitmapWidth+x];
            numSample++;
        }

        bitmapLData[y*bitmapWidth+x]-=bitmapData[y*bitmapWidth+x]*numSample;

    }
}

void Mesh::GetBitmapData(int *& data, int *& ldata, int &w, int &h) {
	data=bitmapData;
	ldata=bitmapLData;
	w=bitmapWidth;
	h=bitmapHeight;
}

void Mesh::MakeSpiralMesh(float rad) {

    m_vVertices.clear();
    m_vFaces.clear();
    triangleBounds.clear();

    int samp=20;

    QList <Vector3> spine=GetSpiralMeshSpine();

    //trace out all the verts
    for (int i=0; i<spine.size(); ++i) {

        Vector3 tanvec;
        if (i==spine.size()-1)
            tanvec=spine[i]-spine[i-1];
        else
            tanvec=spine[i+1]-spine[i];
        tanvec.Normalize();

        Vector3 b1vec=Vector3(0,0,1);
        Vector3 b2vec=tanvec.Cross(b1vec);
        b2vec.Normalize();
        b1vec=tanvec.Cross(b2vec);
        b1vec.Normalize();

        for (float t=0.0f; t<2.0f*M_PI;t+=2.0f*M_PI/float(samp)) {
            Vector3 eachpt=spine[i] + b1vec*cosf(t)*rad + b2vec*sinf(t)*rad;
            m_vVertices.push_back(eachpt);
        }
    }

    //connect all verts to make faces (should be 20 samples per ring)

    for (int i=0; i<spine.size()-1; ++i) {
        for (int j=0; j<samp; ++j) {

            QVector <int> eachface1, eachface2;

            int i1, i2, i3, i4;

            i1=i*samp+j;
            i2=i*samp+(j+1)%samp;
            i3=(i+1)*samp+(j+1)%samp;
            i4=(i+1)*samp+j;

            eachface1.push_back(i1);
            eachface1.push_back(i2);
            eachface1.push_back(i3);

            m_vFaces.push_back(eachface1);

            eachface2.push_back(i1);
            eachface2.push_back(i3);
            eachface2.push_back(i4);

            m_vFaces.push_back(eachface2);
        }
    }

    //add caps
    for (int i=1; i<samp-1; ++i) {

        QVector <int> eachface1, eachface2;

        eachface1.push_back(0);
        eachface1.push_back(i+1);
        eachface1.push_back(i);
        m_vFaces.push_back(eachface1);

        eachface2.push_back((spine.size()-1)*samp+0);
        eachface2.push_back((spine.size()-1)*samp+i);
        eachface2.push_back((spine.size()-1)*samp+i+1);
        m_vFaces.push_back(eachface2);
    }   

    FillInInfo();   

}

QList <Vector3> Mesh::GetSpiralMeshSpine() {

    QList <Vector3> spine;

    for (double l=0.0;l<2.0*M_PI*3.0;l+=M_PI/30.0) {
        Vector3 eachpt;

        eachpt.x=sin(l);
        eachpt.y=cos(l);
        eachpt.z=(l/10.0)*(l/10.0);

        spine.push_back(eachpt);        
    }

    return spine;

}

float Mesh::GetPlaneNeighbourhoodErrorMetric(int v, QVector <QSet <int> > & curNeigh) {

    QVector <Vector3> pts;
    QList <Vector3> eigenvecs;
    QList <float> eigenvals;
    Vector3 centroid;
    float maxdist;

    const QList <int> eachNeigh = curNeigh[v].toList();
    for (int j=0; j<eachNeigh.size(); ++j)
        pts.push_back(m_vVertices[eachNeigh[j]]);

    EigenSolver::PCAForPoints(pts, eigenvecs, eigenvals, centroid, maxdist);

    Vector3 norm = eigenvecs[2];
    norm.Normalize();
    const float d = norm.Dot(centroid);

    //compute error with distance from this plane
    float error = 0.0f;

    for (int j=0; j<eachNeigh.size(); ++j) {
        const Vector3 & eachPt = m_vVertices[eachNeigh[j]];
        error += fabs(norm.Dot(eachPt) - d);
    }

    return error;

}

void Mesh::Decimate(QVector <int> & vSegmentation) {

    qDebug("Vertices before decimation: %i",m_vVertices.size());

    QVector <Vector3> segP = QVector <Vector3> (m_vFaces.size());

    const float oneover3 = 1.0f / 3.0f;

    for (int i=0; i<m_vFaces.size(); ++i) {

        segP[i]=Vector3(0,0,0);

        for (int j=0; j<3; ++j)
            segP[i]+=m_vVertices[m_vFaces[i][j]];

        segP[i]=segP[i] * oneover3;

    }

    //we assume all vertices are simple (each edge shared by exactly 2 tris), this makes things easier
    QVector <float> errorMeasures = QVector <float> (m_vVertices.size());

    //initialize error metrics
    for (int i=0; i<m_vVertices.size(); ++i) {
        if (m_vVertNeigh[i].size() < 3)
            errorMeasures[i] = FLT_MAX;
        else
            errorMeasures[i] = GetPlaneNeighbourhoodErrorMetric(i, m_vVertNeigh);
    }

    //with all errors known, we now remove verts in order of error minimization
    //int vertsToRemove = m_vVertices.size() / 4;
    int vertsToRemove = 0;
    //const float errThreshold = 0.01f;

    QVector <int> vertIndRemoved;

    while (true) {

        int ind = -1;
        float err = FLT_MAX;

        //get index of lowest error vertex
        for (int i=0; i<errorMeasures.size(); ++i) {

            if (errorMeasures[i] < err) {
                ind = i;
                err = errorMeasures[i];
            }

        }

        //if (err > errThreshold || ind < 0)
        if (vertsToRemove > m_vVertices.size() * 8.0 / 10.0f || ind < 0)
            break;

        ++vertsToRemove;
        vertIndRemoved.push_back(ind);
        //qDebug("Still gotta remove %i vertices. Ind: %i Minerror: %f", vertsToRemove, ind, err);

        Vector3 averageNorm = Vector3(0,0,0);

        //remove old faces
        for (int i=0; i<m_vFaces.size(); ++i) {

            if (m_vFaces[i].contains(ind)) {

                averageNorm+=m_vFaceNorms[i];

                m_vFaces.remove(i);
                m_vFaceNorms.remove(i);

                --i;

            }

        }

        averageNorm.Normalize();

        //we have to transition around each vertex of the ring, starting at edge v_0-v_1,
        //and "heal" the hole we just made... my approach is a triangle fan at v_0

        const QList <int> neigh = m_vVertNeigh[ind].toList();
        //qDebug() << "\t" << neigh; //BUG! vertex has its own index in neigh list!!!
        //Still gotta remove 998 vertices. Ind: 2334 Minerror: 0.000000
          //       (2329, 2333, 2334)

        int v0 = neigh.first();

        QList <int> processedNeigh;
        processedNeigh.push_back(v0);

        QSet <int> set;
        set = m_vVertNeigh[ind];
        set.intersect(m_vVertNeigh[v0]);

        QSet <int> v0set = set;
        int v1 = v0set.toList().first(); //TODO: modify this to always go CLOCKWISE (follow face normal)

        QSet <int> v1set = m_vVertNeigh[ind];
        v1set.intersect(m_vVertNeigh[v1]);

        for (int k=0; k<processedNeigh.size(); ++k)
            v1set.remove(processedNeigh[k]);

        int v2 = v1set.toList().first();

        Vector3 vec1 = m_vVertices[v1] - m_vVertices[v0]; vec1.Normalize();
        Vector3 vec2 = m_vVertices[v2] - m_vVertices[v0]; vec2.Normalize();

        bool doClockwise = (vec1.Cross(vec2).Dot(averageNorm) > 0.0f);

        int vend = set.toList().last();

        QVector <QSet <int> > vNewVertNeigh = m_vVertNeigh;

        for (int j=0; j<neigh.size()-2; ++j) {

            set = vNewVertNeigh[ind];
            set.intersect(vNewVertNeigh[v1]);

            for (int k=0; k<processedNeigh.size(); ++k)
                set.remove(processedNeigh[k]);

            int v2 = set.toList().first();

            //qDebug() << "\t" << nextVert;
            //qDebug("\tv0 %i v1 %i v2 %i vend %i vlast %i",v0,v1,v2,vend);
            //if (v0 == v1 || v0 == v2 || v1 == v2)
            //    qDebug() << "\tPROBLEM!" << vNewVertNeigh[ind] << vNewVertNeigh[v1] << nextVert;

            //introduce new face
            QVector <int> newFace;
            newFace.push_back(v0);
            newFace.push_back(v1);
            newFace.push_back(v2);

            m_vFaces.push_back(newFace);

            Vector3 vec1 = m_vVertices[v1] - m_vVertices[v0]; vec1.Normalize();
            Vector3 vec2 = m_vVertices[v2] - m_vVertices[v0]; vec2.Normalize();

            Vector3 faceNorm = vec1.Cross(vec2); faceNorm.Normalize();

            if (!doClockwise)
                faceNorm = faceNorm * (-1.0f);

            m_vFaceNorms.push_back(faceNorm);

            //remove inner vert as neighbour
            vNewVertNeigh[v0].remove(ind);
            vNewVertNeigh[v1].remove(ind);
            vNewVertNeigh[v2].remove(ind);

            //make v0, v2 new neighbours (they are!)
            vNewVertNeigh[v0].insert(v2);
            vNewVertNeigh[v2].insert(v0);

            //move indexes along
            processedNeigh.push_back(v1);
            v1 = v2;

            //if last face to make, end
            if (v2 == vend)
                break;

        }

        m_vVertNeigh = vNewVertNeigh;

        //need to update error metrics
        for (int i=0; i<neigh.size(); ++i) {
            if (errorMeasures[neigh[i]] < FLT_MAX)
                errorMeasures[neigh[i]] = GetPlaneNeighbourhoodErrorMetric(neigh[i], m_vVertNeigh);
        }

        errorMeasures[ind] = FLT_MAX; //this vertex processed

    }

    //actually remove the vertex
    for (int i=0; i<vertIndRemoved.size(); ++i) {

        for (int j=0; j<m_vFaces.size(); ++j) {
            for (int k=0; k<m_vFaces[j].size(); ++k) {
                if (m_vFaces[j][k] >= vertIndRemoved[i]) {
                    --m_vFaces[j][k];
                }
            }
        }

        for (int j=i+1; j<vertIndRemoved.size(); ++j) {
            if (vertIndRemoved[j] >= vertIndRemoved[i])
                --vertIndRemoved[j];
        }

        m_vVertices.remove(vertIndRemoved[i]);
    }

    //compute new segmentation membership
    QVector <int> newSegmentation = QVector <int> (m_vFaces.size());

    for (int i=0; i<m_vFaces.size(); ++i) {

        Vector3 eachP = Vector3(0,0,0);
        for (int j=0; j<3; ++j)
            eachP+=m_vVertices[m_vFaces[i][j]];
        eachP=eachP*(1.0f/3.0f);

        float dist = FLT_MAX;
        int ind = 0;

        for (int j=0; j<segP.size(); ++j) {

            float eachdist = (segP[j] - eachP).Length();

            if (eachdist < dist) {
                dist = eachdist;
                ind = j;
            }

        }

        //qDebug() << "Face " << i << " minDist " << dist << " minind " << ind;
        //qDebug("%f,%f,%f ",eachP.x, eachP.y, eachP.z);

        newSegmentation[i] = vSegmentation[ind];

    }

    vSegmentation = newSegmentation;

    qDebug("Vertices after decimation: %i", m_vVertices.size());

}

float Mesh::GetRandomGaussian(float fMean, float fStd) {

    const int N = 20;
    float x = 0.0f;

    for (int i=0; i<N; ++i)
        x+=float(qrand())/float(RAND_MAX);

    x-= float(N)/2.0f; //adjust mean to 0
    x*= sqrtf(12.0f / float(N)); //adjust variance to 1

    //qDebug("Returning: %f", fMean + fStd * x);
    return fMean + fStd * x; //incorporate parameters for distribution

}

void Mesh::Noisify() {

    const float fMean = 0.0f;
    const float fStd = 0.015f;

    //apply spatial variation to vertices
    for (int i=0; i<m_vVertices.size(); ++i) {

        m_vVertices[i].x += GetRandomGaussian(fMean, fStd);
        m_vVertices[i].y += GetRandomGaussian(fMean, fStd);
        m_vVertices[i].z += GetRandomGaussian(fMean, fStd);

    }

    //recompute normals
    FillInInfo();
    FlipFaceNorms();

}
