// ObjFile.h: interface for the ObjFile class.
//
//////////////////////////////////////////////////////////////////////

#pragma once

#ifndef __GEOMETRY_OBJFILE_H__
#define __GEOMETRY_OBJFILE_H__

#include <QVector>
#include <QtOpenGL>

#include <map>
#include <string>
#include <math.h>
#include <float.h>

#include <string>

#include <stdio.h>
#include <stdarg.h>

#include "Vector.h"

using namespace std;

struct tFace
{
	QVector <int> vertIndex;
	QVector <int> coordIndex;
	QVector <int> normalIndex;
	Vector3 faceNormal;
};

class ObjFile
{
private:

	struct tMaterialInfo
	{
		char			strName[255];		// The texture name
		char			strFile[255];		// The texture file name (If this is set it's a texture map)
		float			color[3];			// The color of the object (R, G, B)
		float			specColor[3];		// The specular color of the object	
		unsigned int	textureId;			// the texture ID
		float			uTile;				// u tiling of texture  (Currently not used)
		float			vTile;				// v tiling of texture	(Currently not used)
		float			uOffset;	 	    // u offset of texture	(Currently not used)
		float			vOffset;			// v offset of texture	(Currently not used)
	};
	
	struct t3DObject
	{
		int  numOfVerts;			// The number of verts in the model
		int  numOfFaces;			// The number of faces in the model
		int  numTexVertex;			// The number of texture coordinates
		int  numNormals;
		int  materialID;			// The texture ID to use, which is the index into our texture array
		char materialName[255];
		bool bHasTexture;			// This is TRUE if there is a texture map for this object
		char strName[255];			// The name of the object
		Vector3  *pVerts;			// The object's vertices
		Vector3  *pNormals;			// The object's normals
		Vector2  *pTexVerts;		// The texture's UV coordinates
		tFace *pFaces;				// The faces information of the object
	};
	
public:
	ObjFile();
	virtual ~ObjFile();

	bool Read(const char *sFilePath);
	bool Read(const char *sFilePath, const char *sBasePath);

	void DrawSimple();	
	void DrawSurface(float specificAlpha=1.0f);
	void DrawWireFrame();	

	//Basic functions
	int getNumObjects();
	tFace *getFaces(int whichObject);
	int getNumFaces(int whichObject);
	int getNumVerts(int whichObject);
	int getNumNorms(int whichObject);
	Vector3 *getVerts(int whichObject);
	Vector3 *getNorms(int whichObject);
	QVector <QVector <int> > getVertNeighbours(int whichObject);

	double GetBoundRad();
	Vector3 GetBoundCentre();
	Vector3 GetBoundLo();
	Vector3 GetBoundHi();


	void DrawRectilinearLines();

	static float ms_defaultDiffuse[4];
	static float ms_defaultWireDiffuse[4];
	static float ms_defaultWireWidth;

private:

	bool FaceIsInside( int& whichObj, tFace * pFace, float lx, float ly, float lz, float width );

	// Obj reading functions
	bool ImportObj();
	void ReadObjFile(FILE* _f);
	void ReadVertexInfo(FILE* _f);
	void ReadFaceInfo(FILE* _f);
	void FillInObjectInfo();
	void ReadMaterialInfo(FILE* _f);
	void ReadMaterialToUse(FILE* _f);

private:
	QVector <tMaterialInfo>	m_vMaterials;		//The list of material information (Textures and colors)
	QVector <t3DObject>		m_vObjects;			//The object list for our model

	bool						m_bObjectHasUV;
	bool						m_bHasTexture;
	bool						m_bObjectHasNormals;
	bool						m_bJustReadAFace;
	int							m_iMaterialID;
	float						m_fBoundingRadius;

	bool m_bObjectFromOffFile;

	//Temp stuff
	QVector <Vector3>		m_vVertices;		//Temp vertices vector
	QVector <tFace>			m_vFaces;			//Temp Faces vector
	QVector <Vector2>		m_vTextureCoords;	//Temp Texture Coords vector
	QVector <Vector3>		m_vNormals;			//Temp Normals vector

	char * basePathStr;
	char * fileNameStr;

	double boundingSphereRad;
	Vector3 boundingSphereCentre;
	Vector3 meshLoBounds;
	Vector3 meshHiBounds;

};

#endif // __GEOMETRY_OBJFILE_H__
