#pragma once

#include <QtOpenGL>

#include <stdio.h>
#include <stdlib.h>

#include <iostream>
#include <sstream>
#include <string>

#include <ctime>

#include "TriMesh.h"
#include "TriMesh_algo.h"

#include "Mesh.h"
#include "ObjFileSimple.h"
#include "OFFFile.h"
#include "PPMFile.h"

extern "C" {
#include "triangle.h"
}

#include "eigensolver.h"
#include "kmeanscluster.h"
#include "binspace.h"

#include "SampleCamera.h"

#define BOUNDRAD_RENDERDISTANCE_FACTOR 1.4f
#define BOUNDRAD_VIEWDISTANCE_FACTOR 2.0f
#define BUFFER_SIZE 256

#define MAX_CURVEPROFILE_SIZE 8

#define NUM_FEATURE_TYPES 10
#define NUM_HUMAN_DATASETS 18

using namespace std;

struct MeshEdge {

    MeshEdge() {
        v1i = -1;
        v2i = -1;
    }

    MeshEdge(int i1, int i2) {
        v1i = i1;
        v2i = i2;
    }

    int v1i;
    int v2i;

    bool operator==(const MeshEdge & e) {        
        return ((v1i == e.v1i) && (v2i == e.v2i)) || ((v1i == e.v2i) && (v2i == e.v1i));
    }

};

struct MeshSegment {  

    Vector3 pt1;
    MeshEdge edge1;

    Vector3 pt2;
    MeshEdge edge2;

    Vector3 n;
    float curv;
    int face;

    void flip() {

        Vector3 tempv;
        MeshEdge tempe;

        tempv = pt1;
        pt1 = pt2;
        pt2 = tempv;

        tempe = edge1;
        edge1 = edge2;
        edge2 = tempe;

    }

};

struct MeshChain {

    MeshChain() {
        isHole=false;
        isClosed=false;
    }

    QList <MeshSegment> segments;
    bool isHole;
    bool isClosed;

    void Draw(float pushOut = 0.0025f) const;
    void Reverse();
    float Length() const;
    Vector3 StartTan() const;
    Vector3 EndTan() const;

};

struct Triangle {
    Vector3 p1, p2, p3;
};

struct Axis {
    Vector3 p;
    Vector3 dir; //normalized!
    float len;
};

struct Plane {		
    Vector3 norm;
    Vector3 rotAxis1;
    Vector3 rotAxis2;
    float pt;
    Vector3 cent;

    Vector3 posB1, posB2, posB3; //ball manip tool positions

    Vector3 featcentroid; //for PCA planes, defines their key point
    float extent; //for PCA planes, defines how far they can move

};

struct MeshFeature {

    MeshFeatureType type;
    int typeindex; //index from list of features of the specific type

    Plane p; //if plane-type feature, this is used
    Axis a; //if axis-type feature, this is used

    bool usedPerpSupport;

};

struct PlanarCut {

    Plane params;
    QList <MeshSegment> segments;
    QList <MeshChain> chains;
    QList <Triangle> tris;

    QList <QList <QList <Vector3> > > annotatePatch;
    QList <Vector3> annotatePatchMidpoint;

    float area;
    int nearSeg;   

    int nVotes;
    float fConfidence;

    PlanarCut() : area(0.0f) { }
    void DrawSegments() const;
    void DrawChains(float pushOut = 0.0025f) const;
    void DrawTrisWireframe() const;
    void DrawTrisFilled() const;

};

struct PuppetCut {

    Vector3 pt1;
    Vector3 pt2;
    int otherPlane;

};

enum PlaneManip {
    NONE,
    ROT1,
    ROT2,
    TRANS,
    NONEBALL,
    MOVEBALL1,
    MOVEBALL2,
    MOVEBALL3
};

class PlaneMesh : public QThread {

    Q_OBJECT

public:

    PlaneMesh(SampleCamera *cam);   

    bool LoadInputMesh(char *filename, char *modelName, char * segName, bool loadHumanData);

    bool ExportContours(char *filename);
    void OutputHTMLTable(std::ostream & os, const QList <QList <int> >  & fData1, const QList <QList <int> >  & fData2, const QList <QString> & rowLabels, const QList <QString> & colLabels);

    bool ExportInputMesh_OBJ(char * filename);
    bool ExportOutputMesh_OBJ(char * filename);
    bool ExportOutputMeshAsGeom_OBJ(char *filename, float ext);
    bool ExportOutputMeshAsSlabs_OBJ(char * filename, float slabThick);
    bool ExportToIllustratorAI(char * filename);
    bool ExportToSVG(char * filename);
    void WriteFeatureGeom(QTextStream & ofs, const Vector3 & p1, const Vector3 & p2, int & vertcount, float ext = 0.01);
    void WriteBoxGeom(QTextStream & ofs, const Vector3 & p1, const Vector3 & p2, int & vertcount);
    bool ExportFeatures_OBJ(char * filename);

    void GetPlanarCut(PlanarCut & planarCut);
    void AddPlanarCut(PlanarCut & planarCut);

    bool AddNewPlane();
    void SetPlaneManipMethod(PlaneManip m);

    float GetBoundRad();
    Vector3 GetBoundCentre();
    Vector3 GetBoundLo();
    Vector3 GetBoundHi();

    void DrawInputVerts();
    void DrawInputWireframe();
    void DrawInputMesh(double meshAlpha);    
    void DrawMeshFeatures();
    void DrawHumanCuts();
    void DrawOutputMesh(bool drawInteriorTriangulation, bool drawContoursInColour);
    void DrawBinSpace();

    void SelectPlane(unsigned int mouse_x, unsigned int mouse_y);
    void SetSelectedPlaneIndex(int i);
    unsigned int GetSelectedPlaneIndex();
    unsigned int GetNumPlanes();        
    float GetPlaneCutArea(int planeIndex);
    Vector3 GetRotatedAxis(const float angle, const Vector3 & vec, const Vector3 & axis);
    float GetPolygonArea(Vector3 v1, Vector3 v2, const QList <MeshSegment> & meshSegments);
    float GetPolygonPerimeter(const QList <MeshSegment> & meshSegments);
    float GetPolygonCurvatureVariation(Vector3 v1, Vector3 v2, const QList <MeshSegment> & meshSegments);

    void UpdatePlaneCuts();
    void PerformPlaneCuts();
    void ClearPlanes();
    void ErasePlane(int planeIndex);  

    bool mousePressEvent(QMouseEvent *event, int winWidth, int winHeight);
    bool mouseReleaseEvent(QMouseEvent *event, int winWidth, int winHeight);
    bool mouseMoveEvent(QMouseEvent *event, int winWidth, int winHeight);   

    void SnapPlaneNormalToXAxis();
    void SnapPlaneNormalToYAxis();
    void SnapPlaneNormalToZAxis();

    void SetUseSegmentations(bool s);
    void SetCubicHermiteInterpSegmentChains(bool b);
    void SetCubicHermiteTanScaleFactor(float d);
    void SetCurvatureSmoothing(double d);
    void SetCurvatureThreshold(double d);
    void SetUseMeshFeatures(bool *use);
    void SetWeightMeshFeatures(float *weights);
    void SetRadiiMeshFeatures(float *radii);
    void SetSelectBinThreshold(float f);
    void SetPlaneParamsFromNormPt(Plane &p);
    void SetPlaneParamsFromNormCent(Plane &p);
    Vector3 SetColorByIndex(const unsigned int index);
    void SetAestheticCuts(bool b, float f);

    void ComputeMeshFeatures(); //computes all features except symmetries (loaded from own file on mesh load)
    void GetSegmentsCovered(const PlanarCut & planarcut, QSet <int> & coveredSegments);
    void LocalSearchFindMaxAreaPlane(PlanarCut & cut, float fRad, const float fRadThreshold);
    void ComputeMatrixA(QString matfilename);

    bool UnprojectPos(QPoint pos, Vector3 & v);
    void UnprojectPosOntoInputMesh(QPoint pos, Vector3 & v);

    void DoAnalysisOnUserData();

    //Applications - annotation
    void AddNewAnnotationPatch();
    void AddToAnnotatePlane(int x, int y, bool newStroke);
    void SelectAnnotatePlane(int x, int y);
    void DrawAnnotations();

    //Applications - navigation
    void GetClosestCurvePointForNav(const Vector3 pt, float & closestLen, Vector3 & closestPt);

    //applications - paper puppets
    void ComputePaperPuppetStuff();
    void DrawPaperPuppetStuff();

    void UpdateBinSpaceDispList();

    void SetCoverAllSegments(bool b);

    //user study analysis (but useful enough to be more than this)
    static QString GetModelFilename(const int mNum);
    static QString GetUserModelFilename(const int uNum, const int mNum);
    static QVector <Plane> LoadUserPlanes(const int uNum, const int mNum);

    QVector <Plane> ComputeSuperproxy(const QVector <QVector <Plane> > & pvecs);

    void GetChainsAs2DPoints(const PlanarCut & planarCut, QList <QList <Vector2> > & points, QVector <Vector2> & loBounds, QVector <Vector2> & hiBounds, Vector2 & loBoundAll, Vector2 & hiBoundAll);    
    void SetBackgroundColour(const QColor & c);
    QColor & GetBackgroundColour();
    void SetMeshLightColour(const QColor & c);

    //do the algorithm
    void run();
    void DoBinningAlgorithm();

public slots:

    void SetDoTriangulation(bool b);
    void SetCheckForHoles(bool b);

    //Applications - robustness
    void MeshDecimate();
    void MeshNoisify();
    void MeshUpdateTrimesh();
    void CutsSegmentify();

    //Web study stuff
    void AddRandomCuts();
    void AddPCAPlanes();

    //User study 1 stuff
    void DoUserStudyAnalysis();

    void FlipMeshNormals();

signals:

    void percentComplete(int);
    void algoFinished();  

private:

    void GetSegments(PlanarCut & planarCut);
    void GetChains(PlanarCut & planarCut);
    void GetTriangles(PlanarCut & planarCut);
    int GetNumChainsOpen(const PlanarCut & planarCut);
    void GetBestFitPlane(const QList <MeshSegment> & points, Plane & plane);
    void GetBestFitAxis(const QList <MeshSegment> & points, Axis & axis);

    void ChainRidgeSegments(PlanarCut & planarCut);

    void RenderPlanePieces(const QList <PlanarCut> & planarCuts, int renderMode);

    void DrawPlane(const Plane & cursorPlane, const Vector3 col);
    void DrawFeaturePlane(const Plane & cursorPlane, const Vector3 col);
    void DrawAxis(const Axis & axis, const Vector3 col);
    void DrawPlaneCoordFrame(const Plane & cursorPlane);

    bool GetPlaneTriIntersect(const Plane & plane, const Triangle & tri, Vector3 & pt1, Vector3 & pt2, int & edge1, int & edge2);
    bool GetRayPlaneIntersect(const Plane & p, const Vector3 & rayP, const Vector3 & rayD, Vector3 & intersectPt);
    bool GetLinePlaneIntersect(const Triangle & tri, const Vector3 & pt1, const Vector3 & pt2, Vector3 &intersectPoint);
    bool GetPlanePlaneIntersect(const Plane & p1, const Plane & p2, Vector3 & ipt, Vector3 & idir);
    bool GetRayTriIntersect(const Vector3 & rayp, const Vector3 & rayd, const Triangle & tri, Vector3 & ipt);
    float GetTriangleArea(const Triangle & t);
    float GetPlaneDistance(const Plane & p1, const Plane & p2);
    float GetPlaneDistance(const Plane & p1, const QVector <Plane> & p2vec);
    float GetPointLineDistance(const Vector3 & p, const Axis & a);

    Vector3 CubicHermiteInterp(const Vector3 & p1, const Vector3 & p2, const Vector3 & t1, const Vector3 & t2, const float t);
    void DoCubicHermiteInterpForChains(MeshChain & chain1, MeshChain & chain2);

    void PiecewisePlanesFromCurve(const QList <MeshSegment> & curvepts, const float planecost, QList <Plane> & bestplanes, QList <QList <MeshSegment> > & pointchains);
    void PiecewiseAxesFromCurve(const QList <MeshSegment> & curvepts, const float planecost, QList <Axis> & bestaxes, QList <QList <MeshSegment> > & pointchains);
    void PiecewiseAxesFromCurve2(const QList <MeshSegment> & curvepts, const float planecost, QList <Axis> & bestaxes, QList <QList <MeshSegment> > & pointchains);

    void RecurseThroughWalkMatrix(int **walkmatrix, int begin, int end, bool *planeend);

    void PCAPlanesForMeshSegment(const int segmentIndex, QList <Vector3> &eigenvecs, QList <float> &eigenvals, Vector3 &centroid);

    void GetFeatures_MeshRidges();
    void GetFeatures_MeshValleys();
    void GetKeyPlanes_PCAForMeshSegments();
    void GetKeyPlanes_GlobalPCA();
    void GetKeyPlanes_GlobalXYZ();

    void MeshRidges(bool do_ridge, float thresh, QList <Vector3> & ridgesegments, QList <Vector3> & ridgenorms, QList <float> & ridgeness);
    void GetFaceRidges(int v0, int v1, int v2, bool do_ridge, float thresh, QList <Vector3> & ridgesegments, QList <Vector3> & ridgenorms, QList <float> & ridgeness);
    void GetSegmentRidge(int v0, int v1, int v2, float emax0, float emax1, float emax2, float kmax0, float kmax1, float kmax2, float thresh, bool to_center,
                         QList <Vector3> & ridgesegments, QList <Vector3> & ridgenorms, QList <float> & ridgeness);

    void PopulateBinSpaceWithFeatures();
    void AddPerpendicularSupport(const Plane & p, const PlaneSpaceBin & maxBin);
    void AddSymmetricSupport(const Plane & p, const PlaneSpaceBin & maxBin);   

    static QList <int> GetRandomSet(const int lo, const int hi, const int size);

    //Class members
public:
    const char *featureTypeStrings[NUM_FEATURE_TYPES];

private:

    Mesh *inputMesh;
    SampleCamera *camera;
    // Globals: mesh...
    char *meshfilename;
    TriMesh *themesh;

    QList <PlanarCut> m_vPlanarCuts;

    PlanarCut curPlanarCut;     

    int selectedPlaneIndex;
    PlaneManip planeManip;    
    QPoint lastPos;

    GLint origViewport[4];
    GLdouble origProj[16];
    GLdouble origModl[16];

    bool m_bDoTriangulation;
    bool m_bCheckForHoles;

    //mesh segmentation related
    int m_nSegments;
    QVector <int> m_vMeshFaceSeg;
    bool m_bUseFaceSeg;

    bool m_bCubicHermiteSegmentInterp;
    int m_bForceUseOneChainOnly;

    float m_fHermiteTanScaleFactor;
    float m_fCurvatureSmoothing;  

    QList <Vector3> distinctColours;
    //TESTING NEW ALGORITHM METHODS       

    //Planes/lines from PCA on segments...    
    QList <Plane> pcaPlaneFits;
    QList <Axis> pcaLineFits;

    //Planes (fitting subchains) from ridges...
    float m_fRidgeValleyThreshold;

    PlanarCut ridgecut;    
    QList <Axis> ridgeAxes;
    QList <QList <MeshSegment> > ridgeAxesChains;
    QList <Vector3> ridgeAxesChainsAvgNorm;
    QList <float> ridgeAxesChainsAvgCurv;
    QList <float> ridgeAxesChainsLength;

    //Planes (fitting subchains) from valleys...
    PlanarCut valleycut;
    QList <Axis> valleyAxes;
    QList <QList <MeshSegment> > valleyAxesChains;
    QList <Vector3> valleyAxesChainsAvgNorm;
    QList <float> valleyAxesChainsAvgCurv;
    QList <float> valleyAxesChainsLength;

    //Planes/lines from symmetry (currently for whole mesh)
    QList <Plane> symPlanes;
    QList <Axis> symLines;

    //Global XYZ/PCA planes
    QList <Plane> globalxyzPlanes;
    QList <Plane> globalpcaPlanes;

    QList <Plane> perpSupportPlanes;
    QList <Plane> symSupportPlanes;

    //FEATURE LIST (unified)
    QList <MeshFeature> meshFeatures;    

    //Binspace
    BinSpace m_BinSpace;
    QVector <PlaneSpaceBin> m_vSelectedBins;

    //binspace params
    float m_fWeights[NUM_FEATURE_TYPES];
    float m_fRadii[NUM_FEATURE_TYPES];
    bool m_bUseFeature[NUM_FEATURE_TYPES];

    //For figure generation
    QList <QList <PlanarCut> > figureCuts;
    QList <Vector3> figureColours;

    QList <PlanarCut> piecewiseFigureCuts;
    QList <QList <Vector3> > piecewiseFigureChains;

    QList <Plane> figureCutsUnselected;

    //user vs model # of planes used figure
    QList <QList <int> > fFigure2Data;
    QList <QString> rowLabelsFigure2;
    QList <QString> colLabelsFigure2;

    //user vs user uniqueness
    QList <QList <int> > fFigure3Data;   

    bool m_bDoCoverAllSegments;

    bool m_baesthetic;
    float m_baestheticthresh;

    QVector <QVector <PuppetCut> > m_puppetcuts;

    QColor m_backColor;
    QColor m_meshLightColor;

};
