/*
	File:	SampleCamera.cpp
	Author:	Azam Khan
	Date:	16dec00

	Usage:	Implements a 3D camera using 3 points, basically
			the parameters to gluLookAt i.e. (eye, center, up).
*/

#include "glew.h"
#include <QtOpenGL>

#include <math.h>
#include <float.h>

#include "Vector.h"
#include "SampleCamera.h"


SampleCamera::SampleCamera( void ) :
	m_vEye	(100,100,100),
	m_vCenter	(0,0,0),
	m_dHorizFov ( 35 ),
	m_dVertFov	( 35 ), // default to a 1:1 aspect ratio
	m_dOrthoWidth ( 4.0 ),
	m_dOrthoHeight ( 4.0 ),
	m_dAspectRatio( 1 ),
	m_bOrthographic( false ),
	m_bPerspWithOrtho( false )

	// PROJECTION_TRANSITION
	// Note: Comment out to test non-implementation of projection transition
	,m_dProjTransition( -1.0 )
	
{
	// calculate an up vector by projecting the world up onto the
	// plane of possible up vectors
	m_vUp				= Vector3(0,1,0);
	Vector3 vViewDir	= (m_vCenter - m_vEye).Normalize();
	Vector3 vRight		= vViewDir.Cross( m_vUp );
	m_vUp				= vRight.Cross( vViewDir ).Normalize();

	m_dBaseOrthoWidth = m_dOrthoWidth;
	m_dBaseOrthoHeight = m_dOrthoHeight;

        nearPlaneDist=0.01;
        farPlaneDist=100.0;

        //GLfloat light_diff[] =		{ 0.9f, 0.9f, 0.9f, 0.8f };
        //static const GLfloat light_spec[] =		{ 1.f, 1.f, 1.f, 1.f };
        //static const GLfloat light_position[] = { 0, 0, 0, 1.f };
}

SampleCamera::SampleCamera (
		double in_eyeX,		double in_eyeY,		double in_eyeZ,
		double in_centerX,	double in_centerY,	double in_centerZ,
		double in_upX,		double in_upY,		double in_upZ ) :
	m_vEye	( in_eyeX,		in_eyeY,	in_eyeZ		),
	m_vCenter	( in_centerX,	in_centerY,	in_centerZ	),
	m_vUp		( in_upX,		in_upY,		in_upZ		),
	m_dHorizFov ( 45 ),
	m_dVertFov	( 45 ), // default to a 1:1 aspect ratio
	m_dOrthoWidth ( 4.0 ),
	m_dOrthoHeight ( 4.0 ),
	m_dAspectRatio( 1 ),
	m_bOrthographic( false ),
	m_bPerspWithOrtho( false )

	// PROJECTION_TRANSITION
	// Note: Comment out to test non-implementation of projection transition
	,m_dProjTransition( -1.0 )
	
{
	m_dBaseOrthoWidth = m_dOrthoWidth;
	m_dBaseOrthoHeight = m_dOrthoHeight;

        nearPlaneDist=0.01;
        farPlaneDist=100.0;
}

const SampleCamera& SampleCamera::operator=( const SampleCamera& rhs )
{
	if ( this != &rhs )
	{
		m_vEye		= rhs.m_vEye;
		m_vCenter	= rhs.m_vCenter;
		m_vUp		= rhs.m_vUp;
		m_dHorizFov = rhs.m_dHorizFov;
		m_dVertFov	= rhs.m_dVertFov;
		m_dOrthoWidth = rhs.m_dOrthoWidth;
		m_dBaseOrthoWidth = rhs.m_dBaseOrthoWidth;
		m_dOrthoHeight = rhs.m_dOrthoHeight;
		m_dBaseOrthoHeight = rhs.m_dBaseOrthoHeight;
		m_bOrthographic = rhs.m_bOrthographic;
		m_bPerspWithOrtho = rhs.m_bPerspWithOrtho;

		// PROJECTION_TRANSITION
		// Note: Comment out to test non-implementation of projection transition
		m_dProjTransition = rhs.m_dProjTransition;
	}

	return *this;
}

double SampleCamera::GetAspectRatio()
{
	return m_dAspectRatio;
}

void SampleCamera::SetAspectRatio( double in_dAspectRatio )
{
	m_dAspectRatio = in_dAspectRatio;

	double dDist	= (m_vCenter - m_vEye).Length();
	double x	= dDist * tan( m_dHorizFov * M_PI / 180.0 );
	double y	= x / m_dAspectRatio;

	m_dVertFov = atan( y / dDist ) * 180 / M_PI;

	if ( m_dBaseOrthoWidth < m_dBaseOrthoHeight )
	{
		m_dOrthoWidth = m_dBaseOrthoWidth;
		m_dOrthoHeight = m_dBaseOrthoWidth / m_dAspectRatio;
	}
	else
	{
		m_dOrthoWidth = m_dBaseOrthoHeight * m_dAspectRatio;
		m_dOrthoHeight = m_dBaseOrthoHeight;
	}
}

void SampleCamera::Draw()
{
	glMatrixMode(GL_PROJECTION);
 	
	/* Calculate between projection and orthographic projection matrices to
	 * facilitate transition animation between the two projection modes.
	 */
	
	//JAMES EDIT - move back frustum far plane (to fit in Earth, using km as units)

	// Perspective projection matrix
	double dPerspProj[16];
	glLoadIdentity();
	//gluPerspective( 2*m_dVertFov, m_dAspectRatio, 0.00001, farPlaneDist );	
	gluPerspective( 2*m_dVertFov, m_dAspectRatio, nearPlaneDist, farPlaneDist );	
	glGetDoublev(GL_PROJECTION_MATRIX, dPerspProj);	

	// Orthographic projection matrix
	double dOrthoProj[16];	
	glLoadIdentity();
	/*
	glOrtho(-0.5f * (float)m_dOrthoWidth, +0.5f * (float)m_dOrthoWidth,
			-0.5f * (float)m_dOrthoHeight, +0.5f * (float)m_dOrthoHeight,
			+0.001f, +500.0f );
	*/
	glOrtho(-0.5f * (float)m_dOrthoWidth, +0.5f * (float)m_dOrthoWidth,
			-0.5f * (float)m_dOrthoHeight, +0.5f * (float)m_dOrthoHeight,
			nearPlaneDist, farPlaneDist );
	glGetDoublev(GL_PROJECTION_MATRIX, dOrthoProj);

	// Interpolated projection matrix
	double dTransProj[16];
	glLoadIdentity();

	// PROJECTION_TRANSITION
	// Note: Uncomment and add breakpoint to test non-implementation of projection transition
	/*if (m_dProjTransition != -1.0)
	{
		bool tmp = false;
		tmp;
	}*/
	
	if (m_dProjTransition >= 0.0 && m_dProjTransition <= 1.0) {
		// Calculated interpolated projection matrix
		double modifier = pow( m_dProjTransition, 1/11.2023 );
		for (int k = 0; k < 16; k++)
		{
			dTransProj[k] = dPerspProj[k] + modifier * ( dOrthoProj[k] - dPerspProj[k] );
		}

		glMultMatrixd(dTransProj);
	}
	else if ( m_bOrthographic )
	{
		// Use orthographic projection
		glMultMatrixd(dOrthoProj);
	}
	else
	{
		// Use perspective projection
		glMultMatrixd(dPerspProj);
	}

	/* UNCOMMENT TO REVERT...
	glLoadIdentity();

	if ( m_bOrthographic )
	{
		glOrtho
		(
			-0.5f * (float)m_dOrthoWidth,
			+0.5f * (float)m_dOrthoWidth,
			-0.5f * (float)m_dOrthoHeight,
			+0.5f * (float)m_dOrthoHeight,
			-500.0f, +500.0f
		);
	}
	else
	{
		gluPerspective( 2*m_dVertFov, m_dAspectRatio, 0.001f, 500.0f );	
	}*/


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

        light_diff[0]=0.9f;
        light_diff[1]=0.9f;
        light_diff[2]=0.9f;
        light_diff[3]=0.8f;

        light_spec[0]=1.0f;
        light_spec[1]=1.0f;
        light_spec[2]=1.0f;
        light_spec[3]=1.0f;

        /*
        light_position[0]=m_vEye.x;
        light_position[1]=m_vEye.y;
        light_position[2]=m_vEye.z;
        light_position[3]=1.0f;
        */
        light_position[0]=0.0f;
        light_position[1]=0.0f;
        light_position[2]=0.0f;
        light_position[3]=1.0f;

	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diff);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_spec);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

	gluLookAt(
		m_vEye.x,		m_vEye.y,		m_vEye.z,
		m_vCenter.x,	m_vCenter.y,	m_vCenter.z,
		m_vUp.x,		m_vUp.y,		m_vUp.z );

}

void SampleCamera::SetCamera( const Vector3 &vEye, const Vector3 &vCent, const Vector3 &vUp ) 
{
	m_vEye		= vEye;
	m_vCenter	= vCent;
	m_vUp		= vUp;	
}

void SampleCamera::GetCamera( Vector3 &vEye, Vector3 &vCent, Vector3 &vUp ) const 
{
	vEye		= m_vEye;
	vCent		= m_vCenter;
	vUp			= m_vUp;
}

void SampleCamera::SetUp(Vector3 vNewUp)
{
	m_vUp	= vNewUp.Normalize();
}

void SampleCamera::SetEye(Vector3 vNewEye)
{
	m_vEye = vNewEye;

	Vector3 vViewDir	= (m_vCenter - m_vEye).Normalize();
	Vector3 vRight		= vViewDir.Cross( m_vUp );
	m_vUp				= vRight.Cross( vViewDir ).Normalize();
}

void SampleCamera::SetCent( Vector3 v) {
    m_vCenter=v;
}

void SampleCamera::SetLoc(Vector3 vLoc)
{
	Vector3 vDiff = vLoc - m_vEye;

	m_vEye = vLoc;
	m_vCenter = m_vCenter + vDiff;

	Vector3 vViewDir	= (m_vCenter - m_vEye).Normalize();
	Vector3 vRight		= vViewDir.Cross( m_vUp );
	m_vUp				= vRight.Cross( vViewDir ).Normalize();
}

void SampleCamera::SetFOV( double horizFOV, double vertFOV )
{
	m_dHorizFov = horizFOV;
	m_dVertFov = vertFOV;
	//m_dAspectRatio = m_dHorizFov / m_dVertFov;
}

void SampleCamera::GetOrthoSize ( double &width, double &height ) const
{
	width = m_dOrthoWidth;
	height = m_dOrthoHeight;

}
void SampleCamera::SetOrthoSize( double width, double height )
{
	m_dBaseOrthoWidth = width;
	m_dBaseOrthoHeight = height;

	if ( width < height )
	{ height = width / m_dAspectRatio; }
	else
	{ width = height * m_dAspectRatio; }

	m_dOrthoWidth = width;
	m_dOrthoHeight = height;
}

//JAMES ADDITION: HOVERCAM SEEMS TO WANT THIS FUNCTION
void SampleCamera::PanCenter( double x, double y )
{
	// First, find the u and v coordinate basis for the pan plane.
	Vector3 u = m_vUp;
	Vector3 c = m_vCenter - m_vEye;
	Vector3 v = u.Cross(c);

	// We have x and y coming in as percentage of movement,
	// so scale it to our current zooming
	x *= 2 * c.Length() * tan(m_dVertFov * M_PI / 180.0);
	y *= 2 * c.Length() * tan(m_dHorizFov * M_PI / 180.0);

	u.Normalize();
	v.Normalize();

	// Now, apply the move as linear combinations of u and v and
	// move back to m_vCenter.
	m_vCenter = m_vCenter + u*x + v*y;
	m_vEye    = m_vEye    + u*x + v*y;
}

void SampleCamera::ZoomTowardCenter( double t )
{
	// In Maya, zoom (linearly) moves the eye toward the center until they
	// are coincident.  Further zooming pushes both points together.

	Vector3 ptDirection = m_vCenter - m_vEye;
	ptDirection.Normalize();

        m_vEye = m_vEye + ptDirection * t;
}

void SampleCamera::Tumble(double x, double y)
{
	double dTheta, dSine, dCosine;
	
	Vector3 vDir	= m_vCenter - m_vEye;
	Vector3 vRight	= vDir.Cross( m_vUp ).Normalize();

	// Translate so that all rotation is about origin
	m_vEye	=	m_vEye - m_vCenter;

	// Take care of the rotation about the horizontal axis (y Movement)
	double d	= sqrt( vRight.y*vRight.y + vRight.z*vRight.z );

	if ( d != 0 )
	{
		dCosine	= vRight.z / d;
		dSine	= vRight.y / d;

		// Project axis of rotation onto xz plane
		m_vEye	=	Vector3( m_vEye.x, dCosine*m_vEye.y - dSine*m_vEye.z, dCosine*m_vEye.z + dSine*m_vEye.y );

		// Rotate axis of rotation to z axis
		m_vEye	=	Vector3( m_vEye.x*d - m_vEye.z*vRight.x, m_vEye.y, m_vEye.x*vRight.x + m_vEye.z*d );
	}

	// Rotate about z axis
	dTheta	= y * M_PI;
	dSine	= sin( dTheta );
	dCosine = cos( dTheta );

	m_vEye	= Vector3( dCosine*m_vEye.x + dSine*m_vEye.y,	-dSine*m_vEye.x + dCosine*m_vEye.y, m_vEye.z  );

	if ( d != 0 )
	{
		dCosine	= vRight.z / d;
		dSine	= vRight.y / d;

		// Undo rotation onto z axis
		m_vEye	= Vector3( m_vEye.x*d + m_vEye.z*vRight.x, m_vEye.y, -m_vEye.x*vRight.x + m_vEye.z*d );

		// Undo rotation onto xz plane
		m_vEye	= Vector3( m_vEye.x, dCosine*m_vEye.y + dSine*m_vEye.z, dCosine*m_vEye.z - dSine*m_vEye.y );
	}

	// Update our up vector
	vDir	= m_vCenter - m_vEye - m_vCenter;
	m_vUp	= vRight.Cross( vDir ).Normalize();

	// Take care of the rotation about the y axis (x movement)
	dTheta	= - x * M_PI;
	dSine	= sin( dTheta );
	dCosine = cos( dTheta );

	m_vEye	= Vector3( dCosine*m_vEye.x + dSine*m_vEye.z,	m_vEye.y, -dSine*m_vEye.x + dCosine*m_vEye.z );
	m_vUp	= Vector3( dCosine*m_vUp.x + dSine*m_vUp.z,	m_vUp.y, -dSine*m_vUp.x + dCosine*m_vUp.z );

	// Inverse the translation above
	m_vEye	= m_vEye + m_vCenter;
}

void SampleCamera::RollUp( double dTheta )
{
	Vector3 vDir	= m_vCenter - m_vEye;
	Vector3 vRight	= vDir.Cross( m_vUp ).Normalize();

	m_vUp			= vRight*(-sin( dTheta )) + m_vUp*cos( dTheta );	
	
	// Just make sure it's still normalized
	m_vUp.Normalize();
}

void SampleCamera::setNearFarPlaneDist(double n, double f) {

	nearPlaneDist=n;
	farPlaneDist=f;

}

void SampleCamera::getNearFarPlaneDist(double &n, double &f) {
	n=nearPlaneDist;
	f=farPlaneDist;
}
