#pragma once

#include <QtCore>
#include <QtOpenGL>
#include <math.h>

#define EPSILON			1e-5

#ifndef M_PI
#define M_PI            3.14159265358979323846  /* pi */
#endif

#ifndef M_PI_2
#define M_PI_2          1.57079632679489661923  /* pi/2 */
#endif

#ifndef M_PI_4
#define M_PI_4          0.78539816339744830962  /* pi/4 */
#endif

#ifndef M_PI_6
#define M_PI_6          0.52359877559829887307  /* pi/6 */
#endif

#ifndef M_1_PI
#define M_1_PI          0.31830988618379067154  /* 1/pi */
#endif

#define ONE_BY_ROOT2	0.70710678118654

#define ROOT_3_OVER_2	0.86602540378444

enum {
	VX = 0,
	VY = 1,
	VZ = 2,
};

class Vector2
{
public:
	// Construction/Destruction
	Vector2()								{};
        Vector2(float _c)						{ x=y=_c; }
        Vector2(float _x, float _y)			{ x=_x; y=_y; }
	Vector2(const Vector2& _rhs)			{ x=_rhs.x; y=_rhs.y; }
//#if defined(_WIN32)
        //explicit Vector2(const CPoint& _rhs)	{ x=static_cast<float>(_rhs.x); y=static_cast<float>(_rhs.y); }
//#endif
	~Vector2() {};

        float Magnitude() { return sqrtf(x*x + y*y); }

	Vector2& operator= (const Vector2& _rhs) { x=_rhs.x; y=_rhs.y; return *this; }
	
	inline Vector2 operator-() const					{ Vector2 _v; _v.x=-x; _v.y=-y; return _v; }
	inline Vector2 operator-(const Vector2& _rhs) const { Vector2 _v; _v.x=x-_rhs.x; _v.y=y-_rhs.y; return _v;}
	inline Vector2 operator+(const Vector2& _rhs) const { return Vector2( x+_rhs.x, y+_rhs.y );	}
        inline Vector2 operator*(const float _rhs) const	{ Vector2 _v; _v.x=x*_rhs; _v.y=y*_rhs; return _v;}
        inline Vector2 operator/(const float _rhs) const	{ Vector2 _v; _v.x=x/_rhs; _v.y=y/_rhs; return _v;}


        Vector2& operator*= (float _c)				{ x*=_c; y*=_c; return *this; }
	Vector2& operator+= (const Vector2& _rhs)	{ x+=_rhs.x; y+=_rhs.y; return *this; }
	Vector2& operator-= (const Vector2& _rhs)	{ x-=_rhs.x; y-=_rhs.y; return *this; }

        inline float Length() const		{ return sqrtf(x*x + y*y); }
        inline const Vector2& Normalize()	{ float len = Length(); x = x/len; y = y/len; return *this;}        

public:
        float x,y;
};

class Vector3
{
public:
	// Construction/Destruction
        Vector3()									{ x=y=z=0.0f; };
        Vector3(float _c)							{ x=y=z=_c; }
        Vector3(float _x, float _y, float _z)	{ x=_x; y=_y; z=_z; }
	Vector3(const Vector3& _rhs)				{ x=_rhs.x; y=_rhs.y; z=_rhs.z; }
	~Vector3() {};

	// Assignment
	const Vector3& operator=(const Vector3& _rhs)		{ x=_rhs.x; y=_rhs.y; z=_rhs.z; return *this; }

	// Operators
	inline Vector3 operator-() const					{ Vector3 _v; _v.x=-x; _v.y=-y; _v.z=-z; return _v; }
	inline Vector3 operator-(const Vector3& _rhs) const { Vector3 _v; _v.x=x-_rhs.x; _v.y=y-_rhs.y; _v.z=z-_rhs.z; return _v;}
	inline Vector3 operator+(const Vector3& _rhs) const { return Vector3( x+_rhs.x, y+_rhs.y, z+_rhs.z );	}

	inline const Vector3& operator+=(const Vector3& _rhs)		{ x+=_rhs.x; y+=_rhs.y; z+=_rhs.z; return *this;}
        inline Vector3 operator*(const float _rhs) const	{ Vector3 _v; _v.x=x*_rhs; _v.y=y*_rhs; _v.z=z*_rhs; return _v;}

	// Indexer
        inline float& operator[](const int i) { return ((float*)this)[i]; }
        inline const float& operator[](const int i) const { return ((float*)this)[i]; }

	// Algebra operators
        inline float Length() const					{ return sqrtf(x*x+y*y+z*z); }
        inline float LengthSquared() const					{ return x*x+y*y+z*z; }
        inline const Vector3& Normalize()				{ float len = Length(); x = x/len; y = y/len; z = z/len; return *this;}
	inline Vector3 Cross(const Vector3& _b)	const	{ return Vector3( y*_b.z-z*_b.y, z*_b.x-x*_b.z, x*_b.y-y*_b.x ); }
        inline float Dot(const Vector3& _b) const		{ return x*_b.x+y*_b.y+z*_b.z; }

	// Relational ops
	inline bool operator==(const Vector3& _rhs) const { return x==_rhs.x && y==_rhs.y && z==_rhs.z; }
	inline bool operator!=(const Vector3& _rhs) const { return x!=_rhs.x || y!=_rhs.y || z!=_rhs.z; }

        // OpenGL
        inline void Draw() const { glVertex3f(x,y,z); }

        inline void Debug(QString name) const { qDebug() << name << " x " << x << " y " << y << " z " << z; }

	static const Vector3 Zero;

public:
        float x,y,z;
};

