#include "binspace.h"

const int BinSpace::m_nMaxTheta = 360;
const int BinSpace::m_nMaxPhi = 181;
const int BinSpace::m_nMaxR = 100;
const int BinSpace::m_nIncr = 2;

BinSpace::BinSpace()
{

    //binspace stuff
    m_DispList=0;

    //binSpace=new PlaneSpaceBin **[m_nMaxTheta];
    binSpace.resize(m_nMaxTheta);

    for (int theta=0; theta<m_nMaxTheta; ++theta) {

        //binSpace[theta]=new PlaneSpaceBin *[m_nMaxPhi];
        binSpace[theta].resize(m_nMaxPhi);

        for (int phi=0; phi<m_nMaxPhi; ++phi) {

            //binSpace[theta][phi]=new PlaneSpaceBin[m_nMaxR];
            binSpace[theta][phi].resize(m_nMaxR);

            for (int r=0; r<m_nMaxR; ++r) {

                binSpace[theta][phi][r].totalvalue=0.0f;
                PolarToCartesian(theta,phi,r,binSpace[theta][phi][r].cartesianN,binSpace[theta][phi][r].cartesianD);

            }

        }

    }

}

void BinSpace::SetSelectThreshold(float f) {

    m_fSelectThreshold = f;

}

float BinSpace::GetSelectThreshold() {

    return m_fSelectThreshold;

}

void BinSpace::AddAxisFeature(Vector3 eachaxis, Vector3 cent, float rad, int fi, int ftype, float value) {

    eachaxis.Normalize();

    const float maxAxisDotProd = cosf( 30.0f * 3.14159f / 180.0f);

    for (int theta=0; theta<m_nMaxTheta; theta+=m_nIncr) {
        for (int phi=0; phi<m_nMaxPhi; phi+=m_nIncr) {
            for (int r=0; r<m_nMaxR; r+=m_nIncr) {

                const Vector3 en=binSpace[theta][phi][r].cartesianN;
                const float ed=binSpace[theta][phi][r].cartesianD;

                if (fabs(en.Dot(eachaxis)) > maxAxisDotProd)
                    continue;

                //here is where we find the closest part on the spine
                Vector3 projn = en - eachaxis * eachaxis.Dot(en);

                if (projn.Length()<0.001f)
                    projn=eachaxis;
                else
                    projn.Normalize();

                float projd=projn.Dot(cent);

                const float len = GetCartesianDistance(projn, projd, en, ed);

                if (len<rad)
                    AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len)/rad);

            }
        }
    }

}

void BinSpace::AddAnisotropicAxisFeature(Vector3 eachaxis, Vector3 cent, float rad, int fi, int ftype, float value, Vector3 anisoDir) {

    eachaxis.Normalize();

    const float maxAxisDotProd = cosf( 30.0f * 3.14159f / 180.0f);

    for (int theta=0; theta<m_nMaxTheta; theta+=m_nIncr) {
        for (int phi=0; phi<m_nMaxPhi; phi+=m_nIncr) {
            for (int r=0; r<m_nMaxR; r+=m_nIncr) {

                const Vector3 & en=binSpace[theta][phi][r].cartesianN;
                const float & ed=binSpace[theta][phi][r].cartesianD;

                if (fabs(en.Dot(eachaxis)) > maxAxisDotProd)
                    continue;

                //here is where we find the closest part on the spine
                Vector3 projn = en - eachaxis * eachaxis.Dot(en);

                if (projn.Length()<0.001f)
                    projn=eachaxis;
                else
                    projn.Normalize();

                float projd=projn.Dot(cent);

                const float len = GetCartesianDistance(projn, projd, en, ed);

                Vector3 otherBin = en * ed;
                Vector3 myBin = anisoDir * anisoDir.Dot(cent);

                Vector3 toBin = otherBin - myBin;
                toBin.Normalize();
                float fAnisoDotProd=toBin.Dot(anisoDir);

                //pointing toward anisodir GOOD
                if (fAnisoDotProd>0.0f) {

                    if (len<rad)
                        AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len)/rad * fAnisoDotProd);

                }


            }
        }
    }

}

void BinSpace::AddCapsuleFeature(Vector3 norm, Vector3 cent, float varyd, float rad, int fi, int ftype, float value) {

    norm.Normalize();
    float pt=norm.Dot(cent);

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                //Doing cartesian space comparison instead
                Vector3 en=binSpace[theta][phi][r].cartesianN;
                float ed=binSpace[theta][phi][r].cartesianD;

                if (fabs(pt)>varyd) { //capsule does not go through d=0 (and hence is one piece)

                    float deltan1=acosf(norm.Dot(en)) * TWO_OVER_PI;
                    float deltad1=fabs(pt-ed)-fabs(varyd);
                    if (deltad1<0.0f)
                        deltad1=0.0f;
                    float len1=(deltan1+deltad1);

                    if (len1<rad)
                        AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len1)/rad);

                }
                else {

                    Vector3 en2=en*-1.0f;
                    float ed2=fabs((pt-fabs(varyd)));

                    float deltan1=acosf(norm.Dot(en)) * TWO_OVER_PI;
                    float deltad1;
                    if (ed<pt+varyd)
                        deltad1=0.0f;
                    else
                        deltad1=ed-(pt+varyd);
                    float len1=(deltan1+deltad1);

                    float deltan2=acosf(norm.Dot(en2)) * TWO_OVER_PI;
                    float deltad2;
                    if (ed<ed2)
                        deltad2=0.0f;
                    else
                        deltad2=ed-ed2;
                    float len2=(deltan2+deltad2);

                    if (len1<rad)
                        AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len1)/rad);
                    else if (len2<rad)
                        AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len2)/rad);

                }


            }
        }
    }   

}

float BinSpace::GetCartesianDistance(const Vector3 & n1, const float & d1, const Vector3 & n2, const float & d2) const {

    const float len1=(acosf(n1.Dot(n2)) * TWO_OVER_PI + fabs(d1-d2));
    const float len2=(acosf((n1 * (-1.0f)).Dot(n2)) * TWO_OVER_PI + fabs(-d1-d2));

    return qMin(len1, len2);

}

void BinSpace::AddAnisotropicPointFeature(Vector3 norm, float d, float rad, int fi, int ftype, float value, Vector3 anisoDir) {

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                const Vector3 en=binSpace[theta][phi][r].cartesianN;
                const float ed=binSpace[theta][phi][r].cartesianD;

                //Function to evaluate for distance in cartesian space, sum of dot product between vectors, and difference of d
                const float len = GetCartesianDistance(norm, d, en, ed);

                Vector3 toBin=en*ed - norm*d;
                toBin.Normalize();
                float fAnisoDotProd=toBin.Dot(anisoDir);

                //pointing toward anisodir GOOD
                if (fAnisoDotProd>0.0f) {

                    if (len<rad)
                        AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len)/rad * fAnisoDotProd);

                }

            }
        }
    } 

}

void BinSpace::AddPointFeature(Vector3 norm, float d, float rad, int fi, int ftype, float value) {

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                const Vector3 en=binSpace[theta][phi][r].cartesianN;
                const float ed=binSpace[theta][phi][r].cartesianD;

                //Function to evaluate for distance in cartesian space, sum of dot product between vectors, and difference of d
                const float len = GetCartesianDistance(norm, d, en, ed);

                if (len < rad)
                    AddFeatureToBin(theta,phi,r,fi,ftype,value*(rad-len)/rad);

            }
        }
    }

}



void BinSpace::AddFeatureToBin(int theta, int phi, int r, int fi, int ftype, float value) {

    BinElement element(fi, ftype, value);
    PlaneSpaceBin & b = binSpace[theta][phi][r];

    if (b.f.contains(element))
        return;

    b.f.push_back(element);
    b.totalvalue+=value;    

    //there is a symmetry where r=0, attempt to add feature there too (this method should recurse no more than once)
    if (r == 0 && phi > 0)
        AddFeatureToBin((theta+180)%360, 180-phi, 0, fi, ftype, value);

}

void BinSpace::Clear() {

    /*
     * Clear bins
     */

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                binSpace[theta][phi][r].f.clear();
                binSpace[theta][phi][r].sf.clear();

                binSpace[theta][phi][r].totalvalue=0.0f;
            }
        }
    }

}

PlaneSpaceBin BinSpace::GetMaxFeaturesBin() {

    int maxt = -1, maxp = -1, maxr = -1;
    int maxfeat=-INT_MAX;
    float maxval=-FLT_MAX;

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                if (binSpace[theta][phi][r].f.size()>maxfeat ||
                    (binSpace[theta][phi][r].f.size() == maxfeat && binSpace[theta][phi][r].totalvalue > maxval) ) {

                    maxfeat=binSpace[theta][phi][r].f.size();
                    maxval=binSpace[theta][phi][r].totalvalue;

                    maxt=theta;
                    maxp=phi;
                    maxr=r;
                }

            }
        }
    }

    qDebug("Maximal featured bin selected at %i,%i,%i with values %i %f",maxt,maxp,maxr,maxfeat,maxval);

    return binSpace[maxt][maxp][maxr];

}

PlaneSpaceBin BinSpace::GetMaxValueBin_IncludeSelected() {

    int maxt = 0, maxp = 0, maxr = 0;
    float maxval = -FLT_MAX;

    for (int theta=0; theta<m_nMaxTheta ;theta+=m_nIncr) {
        for (int phi=0; phi<m_nMaxPhi; phi+=m_nIncr) {
            for (int r=0; r<m_nMaxR; r+=m_nIncr) {

                PlaneSpaceBin & b = binSpace[theta][phi][r];

                if (b.f.empty())
                    continue;

                float eachval = 0.0f;
                bool primaryFeature = false;

                for (int i=0; i<b.f.size(); ++i) {

                    eachval+= b.f[i].value;

                    if (b.f[i].ftype != A_RIDGE && b.f[i].ftype != A_VALLEY)
                         primaryFeature = true;
                }

                if (primaryFeature) {

                    for (int i=0; i<b.sf.size(); ++i) {

                        if (b.sf[i].ftype == A_RIDGE || b.sf[i].ftype == A_VALLEY || b.sf[i].ftype == P_PERPSUPPORT || b.sf[i].ftype == P_SYMSUPPORT)
                            continue;

                        eachval+= b.sf[i].value * 0.5f;
                    }
                }

                if (eachval > maxval) {

                    maxval = eachval;

                    maxt=theta;
                    maxp=phi;
                    maxr=r;
                }

            }
        }
    }

    PlaneSpaceBin & maxbin = binSpace[maxt][maxp][maxr];
    maxbin.totalvalue = maxval;

    qDebug("Max value (unselected+selected) bin at %i,%i,%i (%f,%f,%f,%f) with value %f",
           maxt,maxp,maxr,
           maxbin.cartesianN.x, maxbin.cartesianN.y, maxbin.cartesianN.z, maxbin.cartesianD,
           maxval);

    return binSpace[maxt][maxp][maxr];

}

PlaneSpaceBin BinSpace::GetMaxValueBin() {

    int maxt = 0, maxp = 0, maxr = 0;
    float maxval=-FLT_MAX;

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                if (binSpace[theta][phi][r].totalvalue>maxval) {

                    maxval=binSpace[theta][phi][r].totalvalue;

                    maxt=theta;
                    maxp=phi;
                    maxr=r;
                }

            }
        }
    }

    /*
    qDebug("Max totalvalue bin selected at %i,%i,%i (%f,%f,%f,%f) with value %f",
           maxt,maxp,maxr,
           binSpace[maxt][maxp][maxr].cartesianN.x,binSpace[maxt][maxp][maxr].cartesianN.y,binSpace[maxt][maxp][maxr].cartesianN.z,binSpace[maxt][maxp][maxr].cartesianD,
           maxval);
           */

    return binSpace[maxt][maxp][maxr];

}

PlaneSpaceBin BinSpace::GetNearestPopulatedBin(const int ptheta, const int pphi, const int pr, float & dist) {

    const Vector3 n = binSpace[ptheta][pphi][pr].cartesianN;
    const float d = binSpace[ptheta][pphi][pr].cartesianD;

    dist = FLT_MAX;
    int lt = 0 , lp = 0 , lr = 0;

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                if (!binSpace[theta][phi][r].f.empty()) {

                    const Vector3 en = binSpace[theta][phi][r].cartesianN;
                    const float ed = binSpace[theta][phi][r].cartesianD;

                    float eachDist = GetCartesianDistance(n, d, en, ed);

                    if (eachDist < dist) {

                        dist = eachDist;
                        lt = theta;
                        lp = phi;
                        lr = r;

                    }

                }

            }
        }
    }

    return binSpace[lt][lp][lr];

}

QVector <const PlaneSpaceBin *> BinSpace::GetAllPopulatedBins() {

    QVector <const PlaneSpaceBin *> bins;

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                if (!binSpace[theta][phi][r].f.empty())
                    bins.push_back(&binSpace[theta][phi][r]);

            }
        }
    }

    return bins;

}

void BinSpace::RemoveFeaturesOfMaximalBin(PlaneSpaceBin maxBin) {

    /*
    * Remove features covered by this plane from binspace
    */    

    //qDebug("Removing features belonging to bin (%i of them)!",maxBin.f.size());

    for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            for (int r=0;r<m_nMaxR;r+=m_nIncr) {

                PlaneSpaceBin & eachbin = binSpace[theta][phi][r];

                for (int bi=0; bi<maxBin.f.size(); ++bi) {

                    int i = eachbin.f.indexOf(maxBin.f[bi]);

                    if ( i < 0 )
                        continue;

                    //remove feature (even if slightly covered! this simplifies things a lot)
                    eachbin.totalvalue-=eachbin.f[i].value;
                    eachbin.sf.push_back(eachbin.f[i]);
                    eachbin.f.erase(eachbin.f.begin() + i);

                    if (eachbin.f.empty())
                        eachbin.totalvalue=0.0f;

                }               

            }
        }
    }   

}      

void BinSpace::DrawBin(int theta, int phi, int r) {

    PlaneSpaceBin & bin=binSpace[theta][phi][r];

    //Check if bin contains features...
    if (bin.f.empty() && bin.sf.empty())
        return;

    //If so, draw a cube there...
    const Vector3 v[8]={Vector3(0,0,0),
                      Vector3(0,0,1),
                      Vector3(1,0,1),
                      Vector3(1,0,0),
                      Vector3(0,1,0),
                      Vector3(0,1,1),
                      Vector3(1,1,1),
                      Vector3(1,1,0)};

    //const int inds[24]={0,1,2,3,4,0,3,7,7,3,2,6,6,2,1,5,5,1,0,4,5,4,7,6};
    const int inds[24]={3,2,1,0,
                        7,3,0,4,
                        6,2,3,7,
                        5,1,2,6,
                        4,0,1,5,
                        6,7,4,5};

    int maxindex=-1;
    float maxval=-FLT_MAX;
    float binscale;

    if (!bin.sf.empty()) {

        for (int i=0;i<bin.sf.size();++i) {
            if (bin.sf[i].value>maxval) {
                maxval=bin.sf[i].value;
                maxindex=i;
            }
        }

        binscale=bin.sf[maxindex].value;

    }
    else {

        for (int i=0;i<bin.f.size();++i) {
            if (bin.f[i].value>maxval) {
                maxval=bin.f[i].value;
                maxindex=i;
            }
        }

        binscale=bin.f[maxindex].value;

    }

    binscale*=(float)m_nIncr;

    if (!bin.sf.empty())
        SetColorByIndex(bin.sf[maxindex].ftype);
    else
        glColor3f(1.0f,1.0f,1.0f);

    //draw cube surfaces    
    for (int i=0; i<24; ++i) {

        switch (i) {

        case 0:
            glNormal3f(0,-1,0); break;
        case 4:
            glNormal3f(0,0,-1); break;
        case 8:
            glNormal3f(1,0,0); break;
        case 12:
            glNormal3f(0,0,1); break;
        case 16:
            glNormal3f(-1,0,0); break;
        case 20:
            glNormal3f(0,1,0); break;
        default:
            break;
        }

        glVertex3f(v[inds[i]].x*binscale+theta,
                   v[inds[i]].y*binscale+phi,
                   v[inds[i]].z*binscale+r);
    }  

}

void BinSpace::UpdateDispList() {

    if (m_DispList > 0) {
        glDeleteLists(m_DispList, 1);
        m_DispList = 0;
    }

}

void BinSpace::Draw() {

    glEnable(GL_CULL_FACE);
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

    if (m_DispList>0) {

        glCallList(m_DispList);

    }
    else {

        m_DispList=glGenLists(1);
        glNewList(m_DispList,GL_COMPILE_AND_EXECUTE);

        glDisable(GL_LIGHTING);

        glPushMatrix();

        glScalef(5.0f/360.0f,5.0f/360.0f,5.0f/360.0f);
        glTranslatef(-180.0f,-90.0f,-50.0f);

        glBegin(GL_LINES);

        const int gridLineMajorLine=10;

        for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
            (theta%gridLineMajorLine==0)?glColor3f(0,0,0):glColor3f(0.9,0.9,0.9);
            glVertex3f(theta,0,0);
            glVertex3f(theta,180,0);
        }
        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            (phi%gridLineMajorLine==0)?glColor3f(0,0,0):glColor3f(0.9,0.9,0.9);
            glVertex3f(0,phi,0);
            glVertex3f(360,phi,0);
        }

        for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr) {
            (phi%gridLineMajorLine==0)?glColor3f(0,0,0):glColor3f(0.9,0.9,0.9);
            glVertex3f(0,phi,0);
            glVertex3f(0,phi,100);
        }
        for (int r=0;r<m_nMaxR;r+=m_nIncr) {
            (r%gridLineMajorLine==0)?glColor3f(0,0,0):glColor3f(0.9,0.9,0.9);
            glVertex3f(0,0,r);
            glVertex3f(0,180,r);
        }

        for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr) {
            (theta%gridLineMajorLine==0)?glColor3f(0,0,0):glColor3f(0.9,0.9,0.9);
            glVertex3f(theta,0,0);
            glVertex3f(theta,0,100);
        }
        for (int r=0;r<m_nMaxR;r+=m_nIncr) {
            (r%gridLineMajorLine==0)?glColor3f(0,0,0):glColor3f(0.9,0.9,0.9);
            glVertex3f(0,0,r);
            glVertex3f(360,0,r);
        }

        glEnd();

        glEnable(GL_LIGHTING);
        glBegin(GL_QUADS);

        for (int theta=0;theta<m_nMaxTheta;theta+=m_nIncr)
            for (int phi=0;phi<m_nMaxPhi;phi+=m_nIncr)
                for (int r=0;r<m_nMaxR;r+=m_nIncr)
                    DrawBin(theta,phi,r);

        glEnd();
        glDisable(GL_LIGHTING);

        //draw user cuts (and color according to whether they fall into bins)
        /*
        glDisable(GL_LIGHTING);
        glPointSize(5.0f);
        glColor3f(0,0,0);

        for (int i=0;i<figureCuts.size();i++) {
            for (int j=0;j<figureCuts[i].size();j++) {

                int theta, phi, r;
                CartesianToPolar(figureCuts[i][j].planeParams.norm,figureCuts[i][j].planeParams.pt,theta,phi,r);

                if (theta<=m_nMaxTheta&&phi<=m_nMaxPhi&&r<=m_nMaxR&&binSpace[theta][phi][r].bOncePopulated)
                    glColor3f(0,1,0);
                else
                    glColor3f(1,0,0);

                glBegin(GL_POINTS);
                glVertex3i(theta,phi,r);
                glEnd();

                glBegin(GL_LINES);
                glVertex3i(theta,0,r);
                glVertex3i(theta,phi,r);
                glEnd();

            }
        }
        glColor3f(1,1,1);
        glPointSize(1.0f);
        */

        glPopMatrix();

        glEndList();

    }

    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_LIGHTING);


    //draw coords for selected plane

    if (m_bCursorDraw) {       

        glPushMatrix();

        glScalef(5.0f/360.0f,5.0f/360.0f,5.0f/360.0f);
        glTranslatef(-180.0f,-90.0f,-50.0f);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        glDepthMask(GL_FALSE);

        glLineWidth(5.0f);

        for (int j=0;j<2;j++) {

            if (j==1)
                glDepthFunc(GL_GREATER);

            glBegin(GL_LINES);

            if (j==0)
                glColor4f(1.0f,0.6f,0.6f,1.0f);
            else
                glColor4f(1.0f,0.6f,0.6f,0.5f);

            glVertex3i(0,m_cursorPhi,m_cursorR);
            glVertex3i(std::max(m_nMaxTheta,m_cursorTheta),m_cursorPhi,m_cursorR);

            if (j==0)
                glColor4f(0.6f,1.0f,0.6f,1.0f);
            else
                glColor4f(0.6f,1.0f,0.6f,0.5f);

            glVertex3i(m_cursorTheta,0,m_cursorR);
            glVertex3i(m_cursorTheta,std::max(m_nMaxPhi,m_cursorPhi),m_cursorR);

            if (j==0)
                glColor4f(0.6f,0.6f,1.0f,1.0f);
            else
                glColor4f(0.6f,0.6f,1.0f,0.5f);

            glVertex3i(m_cursorTheta,m_cursorPhi,0);
            glVertex3i(m_cursorTheta,m_cursorPhi,std::max(m_nMaxR,m_cursorR));

            glEnd();

        }

        glDepthMask(GL_TRUE);
        glDepthFunc(GL_LESS);
        glDisable(GL_BLEND);

        glPopMatrix();

        glColor3f(1,1,1);
        glLineWidth(1.0f);

    }   

}

void BinSpace::CartesianToPolarF(Vector3 n, float d, float & theta, float & phi, float & r) {

    theta=atan2(n.z,n.x)*180.0f/(float)M_PI;

    if (theta<0)
        theta+=360.0f;

    phi=acosf(n.y/n.Length())*180.0f/(float)M_PI;

    r=d*100.0f;

    //use plane where r is positive
    if (r<0.0f) {

        theta = theta+180.0f;

        if (theta>=360.0f)
            theta-= 360.0f;

        phi = 180.0f-phi;
        r = -r;
    }

}

void BinSpace::CartesianToPolar(Vector3 n, float d, int & theta, int & phi, int & r) {

    theta=atan2(n.z,n.x)*180.0f/(float)M_PI;
    if (theta<0)
        theta+=360;
    phi=acosf(n.y/n.Length())*180.0f/(float)M_PI;
    r=d*100.0f;

    //use plane where r is positive
    if (r<0) {
        theta=(theta+180)%360;
        phi=180-phi;
        r=-r;
    }

}

void BinSpace::PolarToCartesian(int theta, int phi, int r, Vector3 &n, float &d) {

    const float ftheta=float(theta)/180.0f*float(M_PI);
    const float fphi=float(phi)/180.0f*(float)M_PI;
    const float sinphi=sinf(fphi);

    n.x=cosf(ftheta)*sinphi;
    n.y=cosf(fphi);
    n.z=sinf(ftheta)*sinphi;
    n.Normalize();

    d=float(r)/100.0f;


}

void BinSpace::PolarToCartesianF(float theta, float phi, float r, Vector3 &n, float &d) {

    theta=theta/180.0f*(float)M_PI;
    phi=phi/180.0f*(float)M_PI;

    const float sinphi=sinf(phi);

    n.x=cosf(theta)*sinphi;
    n.y=cosf(phi);
    n.z=sinf(theta)*sinphi;
    n.Normalize();

    d=r/100.0f;

}

bool BinSpace::Empty(int theta, int phi, int r) {

    if (theta<0 || phi<0 || r<0 || theta>=m_nMaxTheta || phi>=m_nMaxPhi || r>=m_nMaxR)
        return false;

    return binSpace[theta][phi][r].f.empty() && binSpace[theta][phi][r].sf.empty();

}

int BinSpace::GetIncr() {

    return m_nIncr;

}

Vector3 BinSpace::SetColorByIndex(int index, float alpha) {

    if (distinctColours.empty()) {

        //for (float shv=1.0f;shv>=0.5f;shv-=0.25f) {

            for (float hue=0.0;hue<360.0f;hue+=30.0f) {
                float v=1.0f;
                float shv=1.0f;
                float c=v*shv;

                float hprime=hue/(60.0f);
                float x=c*(1.0f-fabs(fmod(hprime,2.0f)-1.0f));

                Vector3 eachcol;

                if (hprime<1)
                    eachcol=Vector3(c,x,0);
                else if (hprime<2)
                    eachcol=Vector3(x,c,0);
                else if (hprime<3)
                    eachcol=Vector3(0,c,x);
                else if (hprime<4)
                    eachcol=Vector3(0,x,c);
                else if (hprime<5)
                    eachcol=Vector3(x,0,c);
                else if (hprime<6)
                    eachcol=Vector3(c,0,x);

                float m=v-c;
                eachcol.x+=m;
                eachcol.y+=m;
                eachcol.z+=m;

                distinctColours.push_back(eachcol);
                //qDebug("Colour: %f %f %f",eachcol.x,eachcol.y,eachcol.z);

            }
        //}
    }

    int n=distinctColours.size();

    glColor4f(distinctColours[index%n].x, distinctColours[index%n].y, distinctColours[index%n].z, alpha);
    return distinctColours[index%n];

}

PlaneSpaceBin BinSpace::GetBin(int theta, int phi, int r) {

    Q_ASSERT (theta >= 0);
    Q_ASSERT (phi >= 0);
    Q_ASSERT (r >= 0);
    Q_ASSERT (theta < m_nMaxTheta);
    Q_ASSERT (phi < m_nMaxPhi);
    Q_ASSERT (r < m_nMaxR);

    theta-=theta%m_nIncr;
    phi-=phi%m_nIncr;
    r-=r%m_nIncr;

    return binSpace[theta][phi][r];

}

void BinSpace::SetCursorPos(int t, int p, int r) {

    m_cursorTheta=t;
    m_cursorPhi=p;
    m_cursorR=r;

}

void BinSpace::SetDrawCursor(bool d) {

    m_bCursorDraw=d;

}


